# FILE: colorsphere.tcl
#
# PURPOSE: Writes to stdout an OVF file consisting of two disks, where
#  the 3D vector field in each maps to a different half of the unit
#  sphere S^2. The resulting OVF file can be sent through avf2ppm with a
#  .config file selecting one of the Sat+Hue colormaps to create a color
#  key for use in presentations and publications.

# Run parameters. Adjust as desired.
# NB: To align disk centers with mesh nods, disk_diameter/stepsize
#     should be an odd integer.
set stepsize 1
set disk_diameter 481  ;# Make odd (relative to stepsizsefor center to be a node
set disk_spacing 20
set datafmt "%20.17f"
set linefmt " $datafmt $datafmt $datafmt"

# Setup
# ... Disk centers
set cx [expr {($disk_diameter-$stepsize)/2.}]
set cay [expr {$cx+$disk_diameter+$disk_spacing}] ;# Upper disk
set cby $cx                                       ;# Lower disk
# ... Edge values for lower and upper disks
set j_lowersplit [expr {int(ceil($disk_diameter/double($stepsize)))}]
set j_uppersplit [expr {int(floor(($disk_diameter+$disk_spacing)
                                  /double($stepsize)))}]
set rsq [expr {0.25*$disk_diameter*$disk_diameter}]
set r [expr {sqrt($rsq)}]

set xmin [set ymin [set zmin [expr {-0.5*$stepsize}]]]
set xmax [expr {$xmin+$disk_diameter}]
set ymax [expr {$ymin+2*$disk_diameter+$disk_spacing}]
set zmax [expr {0.5*$stepsize}]

set xbase [set ybase [set zbase 0]]
set inodes [expr {int(round($disk_diameter/double($stepsize)))}]
set jnodes [expr {int(round((2*$disk_diameter+$disk_spacing)/double($stepsize)))}]
set knodes 1


# Write header
puts stdout [subst -nobackslashes -nocommands \
{# OOMMF OVF 2.0
#
# Segment count: 1
#
# Begin: Segment
# Begin: Header
#
# Title: colorsphere.ovf
# Desc: OVF file for making color key for spheres
# meshunit: unknown
# meshtype: rectangular
# xbase: $xbase
# ybase: $ybase
# zbase: $zbase
# xnodes: $inodes
# ynodes: $jnodes
# znodes: $knodes
# xstepsize: $stepsize
# ystepsize: $stepsize
# zstepsize: $stepsize
# xmin: $xmin
# ymin: $ymin
# zmin: $zmin
# xmax: $xmax
# ymax: $ymax
# zmax: $zmax
# valuedim: 3
# valuelabels: mx my mz
# valueunits: {} {} {}
#
# End: Header
#
# Begin: Data Text}]


# Write data
for {set k 0} {$k<$knodes} {incr k} {
   for {set j 0} {$j<$jnodes} {incr j} {
      set y [expr {$ybase+$j*$stepsize}]
      if {$j<$j_lowersplit} { ;# Lower (b) disk
         set ydiff [expr {$y-$cby}]
         set zsign -1
      } elseif {$j<$j_uppersplit} { ;# Disk spacer region
         for {set i 0} {$i<$inodes} {incr i} {
            puts stdout [format $linefmt 0. 0. 0.]
         }
         continue
      } else { ;# Upper (a) disk
         set ydiff [expr {$y-$cay}]
         set zsign 1
      }
      set ydiffsq [expr {$ydiff*$ydiff}]
      set my [expr {$ydiff/$r}]
      for {set i 0} {$i<$inodes} {incr i} {
         set x [expr {$xbase+$i*$stepsize}]
         set xdiff [expr {$x-$cx}]
         set xdiffsq [expr {$xdiff*$xdiff}]
         if {$xdiffsq + $ydiffsq > $rsq} {
            puts stdout [format $linefmt 0. 0. 0.]
         } else {
            set mx [expr {$xdiff/$r}]
            set tmp [expr {1.0-$mx*$mx-$my*$my}]
            set mz [expr {$tmp<=0 ? 0.0 : $zsign*sqrt($tmp)}]
            puts stdout [format $linefmt $mx $my $mz]
         }
      }
   }
}

# Write trailer
puts stdout \
{# End: Data Text
# End: Segment
}
