/* FILE: psdraw.h                 -*-Mode: c++-*-
 *
 * OOMMF C++ classes for PostScript rendering
 *
 */

#ifndef PSDRAW_H
#define PSDRAW_H

#include "colormap.h"
#include "planepoint.h"

/* End includes */     /* Optional directive to build.tcl */

#ifndef PSDRAW_ALT_ARROWS
# define PSDRAW_ALT_ARROWS 0
#endif

//////////////////////////////////////////////////////////////////////////
// OOMMF vector graphics drawing base class
class OommfDrawVectorGraphics {
public:
  OommfDrawVectorGraphics
  (Tcl_Channel channel,
   OC_REAL8m pxoff,OC_REAL8m pyoff,
   OC_REAL8m printwidth,OC_REAL8m printheight,
   const char* page_orientation,
   OC_REAL8m xmin,OC_REAL8m ymin,
   OC_REAL8m xmax,OC_REAL8m ymax,
   OommfPackedRGB background_color,
   OC_REAL8m matwidth,
   OommfPackedRGB matcolor,
   OommfPackedRGB arrow_outline_color,
   OC_REAL8m arrow_outline_width=0.0,
   OC_REAL8m arrow_width     =0.075,
   OC_REAL8m arrow_tip_width =0.01875,
   OC_REAL8m arrow_length    =0.9,
   OC_REAL8m arrow_head_width=0.4,
   OC_REAL8m arrow_head_ilen =0.4,
   OC_REAL8m arrow_head_olen =0.5);
  /// Write header.  Offsets and print sizes are in points, x/y
  /// range values in pixels.  px/yoff are to lower lefthand
  /// corner of image.  Pixel coordinates run from upper lefthand
  /// corner of image.

  virtual ~OommfDrawVectorGraphics() {
     // Child clases should write their file trailer to "channel"
    channel=nullptr;  // Disable channel access
  }

  // Draws (potentially fat) line segments between each successive pair
  // of points in vlist.  If joinstyle == 0, then successive segments
  // are connected with a mitered join, == 1 gives a rounded joint, and
  // == 2 gives a beveled join.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those ends
  // will be joined using joinstyle.
  virtual void DrawPolyLine(const Nb_List<PlanePoint>& vlist,
                            OC_REAL8m linescale,OommfPackedRGB color,
                            int joinstyle) = 0;

  virtual void DrawFilledRectangle(OC_REAL8m x1,OC_REAL8m y1,
                                   OC_REAL8m x2,OC_REAL8m y2,
                                   OommfPackedRGB color,
                                   OC_REAL8m opaque) = 0;
  /// Opposing vertices (x1,y1) and (x2,y2) are included.
  /// Specify transparency_stipple to NULL or "" to get no stipple.

  virtual void DrawFilledArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                               OC_REAL8m xcos,OC_REAL8m ycos,OC_REAL8m zcos,
                               OommfPackedRGB color) = 0;
  /// Draws arrow centered at point (xc,yc) in bitmap, of specified size,
  /// with directional cosines xcos,ycos,zcos, in the specified color.


  virtual void DrawOutofPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                                   OommfPackedRGB outercolor,
                                   OommfPackedRGB innercolor) = 0;
  /// Symbol representing out-of-plane vectors


  virtual void DrawIntoPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                                  OommfPackedRGB outercolor,
                                  OommfPackedRGB innercolor) = 0;
  /// Symbol representing in-to-plane vectors.

  virtual void ApplyMat() = 0;
  /// Lays flat mat around border of specified width and color.

protected:
  Tcl_Channel channel;
  const int bbox_xmin,bbox_ymin,bbox_xmax,bbox_ymax;
  const OC_REAL8m scale;  // Multiply by scale to convert to print units
  const OC_REAL8m mat_width;
  const OommfPackedRGB mat_color;
  const OommfPackedRGB ArrowOutlineColor;
  const OC_REAL8m ArrowOutlineWidth;
  const OC_REAL8m ArrowWidth;
  const OC_REAL8m ArrowTipWidth;
  const OC_REAL8m ArrowLength;
  const OC_REAL8m ArrowHeadWidth;
  const OC_REAL8m ArrowHeadInnerLength;
  const OC_REAL8m ArrowHeadOuterLength;
  const OC_REAL8m InOutTipRadius;

  // Might want to add support for "Seascape" orientation?
  enum class PageOrientation { Portrait, Landscape };
  const PageOrientation page_orientation;

  // Disable copy constructor and assignment operators
  OommfDrawVectorGraphics(const OommfDrawVectorGraphics&) = delete;
  OommfDrawVectorGraphics& operator=(const OommfDrawVectorGraphics&) = delete;
};


//////////////////////////////////////////////////////////////////////////
// OOMMF PostScript drawing class
class OommfPSDraw : public OommfDrawVectorGraphics {
public:
  OommfPSDraw(Tcl_Channel channel,
	      OC_REAL8m pxoff,OC_REAL8m pyoff,
	      OC_REAL8m printwidth,OC_REAL8m printheight,
	      const char* page_orientation,
	      OC_REAL8m xmin,OC_REAL8m ymin,
	      OC_REAL8m xmax,OC_REAL8m ymax,
	      OommfPackedRGB background_color,
              OC_REAL8m matwidth,
              OommfPackedRGB matcolor,
              OommfPackedRGB arrow_outline_color,
              OC_REAL8m arrow_outline_width,
              OC_REAL8m arrow_width,
              OC_REAL8m arrow_tip_width,
              OC_REAL8m arrow_length,
              OC_REAL8m arrow_head_width,
              OC_REAL8m arrow_head_ilen,
              OC_REAL8m arrow_head_olen);
  /// Write header.  Offsets and print sizes are in points, x/y
  /// range values in pixels.  px/yoff are to lower lefthand
  /// corner of image.  Pixel coordinates run from upper lefthand
  /// corner of image.

  ~OommfPSDraw();                   // Write trailer

  // Draws (potentially fat) line segments between each successive pair
  // of points in vlist.  If joinstyle == 0, then successive segments
  // are connected with a mitered join, == 1 gives a rounded joint, and
  // == 2 gives a beveled join.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those ends
  // will be joined using joinstyle.
  void DrawPolyLine(const Nb_List<PlanePoint>& vlist,
		    OC_REAL8m linescale,OommfPackedRGB color,
                    int joinstyle);

  void DrawFilledRectangle(OC_REAL8m x1,OC_REAL8m y1,
			   OC_REAL8m x2,OC_REAL8m y2,
			   OommfPackedRGB color,
                           OC_REAL8m opaque);
  /// Opposing vertices (x1,y1) and (x2,y2) are included.  Opacity is
  /// handled by stipple pattern. Set opaque to 1 to get no stipple.

  void DrawFilledArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
		       OC_REAL8m xcos,OC_REAL8m ycos,OC_REAL8m zcos,
		       OommfPackedRGB color);
  /// Draws arrow centered at point (xc,yc) in bitmap, of specified size,
  /// with directional cosines xcos,ycos,zcos, in the specified color.

  // Symbol representing out-of-plane vectors
  void DrawDiamondWithCross(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
			    OommfPackedRGB outercolor,
			    OommfPackedRGB innercolor);
  void DrawOutofPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                           OommfPackedRGB outercolor,
                           OommfPackedRGB innercolor) {
    DrawDiamondWithCross(xc,yc,size,outercolor,innercolor);
  }

  // Symbol representing in-to-plane vectors.
#if PSDRAW_ALT_ARROWS
  void DrawSquareWithDot(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
			 OommfPackedRGB outercolor,
			 OommfPackedRGB innercolor);
  void DrawIntoPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                          OommfPackedRGB outercolor,
                          OommfPackedRGB innercolor) {
    DrawSquareWithDot(xc,yc,size,outercolor,innercolor);
  }
#else // !PSDRAW_ALT_ARROWS
  void DrawDiskWithDot(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                       OommfPackedRGB outercolor,
                       OommfPackedRGB innercolor);
  void DrawIntoPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                          OommfPackedRGB outercolor,
                          OommfPackedRGB innercolor) {
    DrawDiskWithDot(xc,yc,size,outercolor,innercolor);
  }
#endif // PSDRAW_ALT_ARROWS

  void ApplyMat();
  /// Lays flat mat around border of specified width and color.

  // Disable copy constructor and assignment operators.
  OommfPSDraw(const OommfPSDraw&) = delete;
  OommfPSDraw& operator=(const OommfPSDraw&) = delete;

private:
  static const ClassDoc class_doc;
};

//////////////////////////////////////////////////////////////////////////
// OOMMF SVG drawing class
class OommfSVGDraw : public OommfDrawVectorGraphics {
public:
  OommfSVGDraw(Tcl_Channel channel,
               OC_REAL8m pxoff,OC_REAL8m pyoff,
               OC_REAL8m printwidth,OC_REAL8m printheight,
               const char* page_orientation,
               OC_REAL8m xmin,OC_REAL8m ymin,
               OC_REAL8m xmax,OC_REAL8m ymax,
               OommfPackedRGB background_color,
               OC_REAL8m matwidth,
               OommfPackedRGB matcolor,
               OommfPackedRGB arrow_outline_color,
               OC_REAL8m arrow_outline_width,
               OC_REAL8m arrow_width,
               OC_REAL8m arrow_tip_width,
               OC_REAL8m arrow_length,
               OC_REAL8m arrow_head_width,
               OC_REAL8m arrow_head_ilen,
               OC_REAL8m arrow_head_olen);
  /// Write header.  Offsets and print sizes are in points, x/y
  /// range values in pixels.  px/yoff are to lower lefthand
  /// corner of image.  Pixel coordinates run from upper lefthand
  /// corner of image.

  ~OommfSVGDraw();                   // Write trailer

  // Draws (potentially fat) line segments between each successive pair
  // of points in vlist.  If joinstyle == 0, then successive segments
  // are connected with a mitered join, == 1 gives a rounded joint, and
  // == 2 gives a beveled join.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those ends
  // will be joined using joinstyle.
  void DrawPolyLine(const Nb_List<PlanePoint>& vlist,
		    OC_REAL8m linescale,OommfPackedRGB color,
                    int joinstyle);

  void DrawFilledRectangle(OC_REAL8m x1,OC_REAL8m y1,
			   OC_REAL8m x2,OC_REAL8m y2,
			   OommfPackedRGB color,
                           OC_REAL8m opaque);
  /// Opposing vertices (x1,y1) and (x2,y2) are included.
  /// Opaque == 1 is fully opaque, 0 is fully transparent


  // Symbol for "mostly" in-plane vectors
  void DrawFilledArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
		       OC_REAL8m xcos,OC_REAL8m ycos,OC_REAL8m zcos,
		       OommfPackedRGB color);
  /// Draws arrow centered at point (xc,yc) in bitmap, of specified size,
  /// with directional cosines xcos,ycos,zcos, in the specified color.


  // Symbol representing out-of-plane vectors
  void DrawDiamondWithCross(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
			    OommfPackedRGB outercolor,
			    OommfPackedRGB innercolor);
  void DrawOutofPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                           OommfPackedRGB outercolor,
                           OommfPackedRGB innercolor) {
    DrawDiamondWithCross(xc,yc,size,outercolor,innercolor);
  }

  // Symbol representing in-to-plane vectors.
#if PSDRAW_ALT_ARROWS
  void DrawSquareWithDot(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
			 OommfPackedRGB outercolor,
			 OommfPackedRGB innercolor);
  void DrawIntoPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                          OommfPackedRGB outercolor,
                          OommfPackedRGB innercolor) {
    DrawSquareWithDot(xc,yc,size,outercolor,innercolor);
  }
#else // !PSDRAW_ALT_ARROWS
  void DrawDiskWithDot(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                       OommfPackedRGB outercolor,
                       OommfPackedRGB innercolor);
  void DrawIntoPlaneArrow(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
                          OommfPackedRGB outercolor,
                          OommfPackedRGB innercolor) {
    DrawDiskWithDot(xc,yc,size,outercolor,innercolor);
  }
#endif // PSDRAW_ALT_ARROWS

  void ApplyMat();
  /// Lays flat mat around border of specified width and color.

  // Disable copy constructor and assignment operators.
  OommfSVGDraw(const OommfPSDraw&) = delete;
  OommfSVGDraw& operator=(const OommfPSDraw&) = delete;

private:
  static const ClassDoc class_doc;
};


#endif // PSDRAW_H
