# Tcl support script for makerule.tcl
# This script supports parallel builds off the event loop.
# Launch this from the MakeRule class via
#
#   proc OpenCmdChan { index } {
#      set cmdfile [file join [file dirname [info script]] buildrule.tcl]
#      set chan \
#         [open "| tclsh $cmdfile $index [array get options] 2>@ stderr" r+]
#      fconfigure $chan -blocking 0 -buffering full -buffersize 1000000
#      fileevent $chan readable "$class ReadCmdResultHandler $index $chan"
#      return $chan
#   }
#
# where "index" is a numeric index into the reference list command
# channels in the client application.

set failsafe_timeout 30000 ;# Kill process if it doesn't receive any
## build requests for this long a period (in milliseconds). The
## parent pimake process will automatically restart process channels
## as appropriate.

# Uncomment the 'set DBGFILE' to enable debug writes. User is responsible
# for deleting the (potentially large number of) resulting debug files.
# set DBGFILE "debug-buildrule-[pid].txt"
set reference_index -1
if {[info exists DBGFILE]} {
   proc DBGMSG { msg } {
      global DBGCHAN reference_index
      set ticks [clock milliseconds]
      set secs [expr {$ticks/1000}]
      set ms [expr {$ticks - $secs*1000}]
      puts $DBGCHAN [format "<%s %s.%03d CmdChan %2d> %s" [pid] \
                        [clock format $secs -format {%Y-%b-%d %T}] $ms \
                        $reference_index $msg]
      flush $DBGCHAN
   }
   set DBGCHAN [open $DBGFILE a]
   puts $DBGCHAN "---"
} else {
   proc DBGMSG { msg } {}
}

global env
if {[catch {
   set argv [lassign $argv reference_index] ;# Reference index in client
   DBGMSG "START"
   # Remainder of argv are option value pairs
   set opts [dict create {*}$argv]
   if {[dict exists $opts -verbose]} {
      set verbose [dict get $opts -verbose]
   } else {
      set verbose 1
   }
   if {[dict exists $opts -initscript]} {
      eval [dict get $opts -initscript]
   }

   set env(OOMMF_BUILD_ENVIRONMENT_NEEDED) 1
   set pimake_directory [file dirname [info script]]
   set oommf_directory [file dirname [file dirname $pimake_directory]]
   lappend auto_path [file join $oommf_directory pkg/oc]
   #package require Oc
   source [file join $pimake_directory makerule.tcl]
   source [file join $pimake_directory csourcefile.tcl]
   source [file join $pimake_directory platform.tcl]
   source [file join $pimake_directory procs.tcl]
   MakeRule SetOptions $argv
} errmsg erropts]} {
   DBGMSG "buildrule.tcl initialization error: $errmsg"
   return -options $erropts $errmsg
}

# Write command lines to stdout before executing them
Oc_Exec SetOutChannel stdout

proc WriteMessageHeader { chan msgid } {
   puts $chan "MESSAGE START ID $msgid"
   DBGMSG "MESSAGE START ID $msgid"
}
proc WriteMessageBody { chan msg } {
   puts $chan $msg
   DBGMSG $msg
}
proc WriteMessageTrailer { chan msgid msgcode} {
   puts $chan "ID $msgid RESULT $msgcode MESSAGE END"
   DBGMSG "ID $msgid RESULT $msgcode MESSAGE END"
   after idle [list catch [list flush $chan]]
   # Without the # 'after' the flush may block.
}

# Return from ReadMessage is a two element list. The first
# element a result code and the second is the import buffer
# appended with input from chan. The result codes are
#    0 => Success. In this case the second element is the msg
#         with the message header and trailer stripped off.
#    1 => Partial message; pass buffer back on next readable event
#    2 => EOF on channel
#    3 => Mangled message
proc ReadMessage { chan buffer } {
   set buf [read $chan]
   # Note: Parent won't send a subsequent command until the child
   # replies, so we can read chan all in one go w/o having to worry
   # about separating adjacent messages. The situation is different in
   # the parent Read proc because the child may send a child timeout
   # message at any time.
   if {[catch {eof $chan} status] || $status} {
      catch {close $chan}
      return [list 2 {}]
   }
   append buffer $buf
   set endpt [string first "MESSAGE END\n" $buffer]
   if {$endpt<0} {
      # Partial message. Client should call back when channel is
      # readable again.
      return [list 1 $buffer]
   }
   set buf [string trim $buffer]
   if {![regexp {^MESSAGE START ID (\S+)\n(.*)$} $buf _ msgid tail] \
          || ![regexp "^(.*)\nID $msgid MESSAGE END" $tail _ msg]} {
      return [list 3 $buffer]
   }
   return [list 0 $msg]
}

proc ConfigureCmdChannel { chan {handler {}}} {
   fconfigure $chan -blocking 0 -buffering full -buffersize 1000000
   if {![string match {} $handler]} {
      fileevent $chan readable "$handler $chan"
   }
}

set inchan stdin
set outchan stdout

ConfigureCmdChannel $inchan ReadInput
ConfigureCmdChannel $outchan

puts $outchan "\nChild Ready\n" ;# Report initialization complete
## The mother process ignores all data on $outchan preceding the
## "\nChild Ready\n" message. This way anything dumped to
## stdout/stderr by the buildrule.tcl initialization process is
## irrelevant.

set read_buffer {}
proc ReadInput { inchan } {
   lassign [ReadMessage $inchan $::read_buffer] code ::read_buffer
   if {$code == 1} {
      return  ;# Incomplete message
   } elseif {$code == 2} {
      catch {close $outchan}
      DBGMSG "Normal exit"
      exit    ;# EOF;  This is the normal exit path
   } elseif {$code == 3} {
      error "Mangled message: $::read_buffer"
   }

   # Raise an error if input does not look like a complete command
   if {![info complete $::read_buffer]} {
      error "Mangled command: $::read_buffer"
   }

   # Input looks good. Clear buffer and run command
   set cmd $::read_buffer
   set ::read_buffer {}
   RunCommand $cmd
}

proc RunCommand { cmd } {
   # Note: This command blocks while cmd is running
   if {[info exists ::failsafe_timer_id]} {
      after cancel $::failsafe_timer_id
      unset ::failsafe_timer_id
   }
   DBGMSG "Run $cmd"
   set msgid [expr {int(rand()*pow(2,53))}]
   if {[catch {
      WriteMessageHeader $::outchan $msgid
      set taskcode [catch {eval $cmd} cmdresult erropts]
      if {$taskcode != 0} {
         # Taskcode is reserved for future expansion
         WriteMessageBody $::outchan \
            "*** Pimake build programming error: $cmdresult"
         set code -1
      } else {
         lassign $cmdresult target index cmdcode cmdmsg errinfo
         # Note: Any writes to stdout or stderr from cmd
         # appear in client as part of the message body.
         WriteMessageBody $::outchan $cmdmsg
         set code $cmdcode
         if {$errinfo ne "" && $::verbose>2} {
            WriteMessageBody $::outchan "\nCMD: $cmd\nStack info: $errinfo"
         }
      }
      WriteMessageTrailer $::outchan $msgid $code
   } communication_error]} {
      puts stderr "ERROR: Child communication error, msg $msgid"
      exit 99
   }
   DBGMSG "Cmd complete"
   set ::failsafe_timer_id \
      [after $::failsafe_timeout [list set ::forever -1]]
}

proc SignOut {} {
   # Close input channel and send disconnect message to parent
   # Note: Flush outchan but don't explicitly close it; the flush in
   # WriteMessageTrailer is an 'after' command to be certain it
   # doesn't block, but we want to make sure that the flush happens
   # before the channel is closed.
   catch {fileevent $::inchan readable {}}
   catch {close $::inchan}
   set code 0 ;# Use normal processing (i.e., not an error)
   set msgid [expr {int(rand()*pow(2,53))}]
   WriteMessageHeader $::outchan $msgid
   WriteMessageBody $::outchan "CHILD TIMEOUT [pid]"
   WriteMessageTrailer $::outchan $msgid $code
}

set ::failsafe_timer_id \
   [after $::failsafe_timeout [list set ::forever -1]]

vwait forever

if {$forever == -1} {
   # failsafe timeout
   SignOut
   DBGMSG "Error: Pimake buildrule process [pid] timeout"
   set wait_to_close 0 ;# Make sure I/O can finish
   after 1000 [list set wait_to_close 1]
   vwait wait_to_close
   exit 86
}
DBGMSG "Error: Pimake buildrule process [pid] irregular exit (code $forever)"
exit 13
