# FILE: makerule.tcl
#
# The MakeRule class -- instances of this class are created by pimake
# by evaluating makerules.tcl files.  All the real work of pimake is
# performed by the methods and procs of this class.
#
# DOME: Fix value of $rulesFile when processing default rules.
#       Perhaps reduce calls to Oc_DirectPathname within proc Define
#       by intelligently tracking what rulesFile is being sourced.
########################################################################

Oc_Class MakeRule {
   # The offset between the tickers of the Tcl commands
   # 'file' and 'clock'
   const common offset

   const common defaultRules

   array common options

   common pfx ""

   proc SetOptions {optlist} {
      array set options $optlist
   }

   common numErrors

   proc NumErrors {} {
      return $numErrors
   }

   common errchan stderr

   proc SetErrChannel { chan } {
      set errchan $chan
   }

   ClassConstructor {
      Oc_TempFile New f
      set offset [expr {[file mtime [$f AbsoluteName]] \
                           - [clock seconds]}]
      $f Delete
      set drfile [file join [file dirname [info script]] \
                     makerules.config]
      if {![file readable $drfile] \
             || [catch {
                set f [open $drfile r]
                set defaultRules [read $f]
                close $f
             }]} {
         set defaultRules {}
      }
      set numErrors 0
   }

   # A map from a target to the rule which updates it
   array common ruleWhichMakes

   # Keep track of which makerules files have been read
   array common filesRead

   # Keep track of which targets have been built
   array common built

   # The default target to build
   common defaultTarget
   proc SetDefaultTarget { target } {
      if {[file pathtype $target] ne "absolute"} {
         # target should be an absolute path. If not,
         # prepend with current working directory.
         set target [file join [pwd] $target]
      }
      if {![info exists defaultTarget] || $target eq $defaultTarget} {
         return   ;# NOP
      }
      if {[info exists defaultTarget]} {
         Oc_Log Log "Replacing defaultTarget $defaultTarget with $target" \
            warning $class
      }
      set defaultTarget $target
   }

   # This proc builds the requested target (in direct pathname form),
   # and returns the time at which the target was built.
   proc Build {{target {}}} {
      if {[string match {} $target]} {
         set target $defaultTarget
      }
      if {[info exists options(-d)]} {
         puts stdout "${pfx}Target: $target"
      }
      if {[info exists built($target)]} {
         if {[string match pending $built($target)]} {
            Oc_Log Log "Circular dependency:\n\t'$target'\
                        (ultimately) depends on itself.\n\tDropping\
                        that dependency." warning $class
            # Didn't build target, so don't make a timestamp.
            # Report back that this target is very old so that it
            # won't trigger other updates.
            return 0
         }
         # The target has already been built.
         # Return its update time.
         return $built($target)
      }
      # The target hasn't been built yet.
      # Find the rule for how to build it.
      global errorInfo errorCode
      if {[catch {$class FindRule $target} rule]} {
         foreach {ei ec} [list $errorInfo $errorCode] {}
         set emsg $rule
         set prune "\n    ('$class' proc 'FindRule'"
      }
      if {![info exists emsg] && [string match {} $rule]} {
         # No rule to make this target.  See if it's a file.
         # NB: Each MakeRule Define statement in a sourced makerules.tcl
         #     file creates a rule. If the Define does not specify
         #     -script then the default script is {}.
         if {[file exists $target]} {
            # Save and return the file's timestamp
            # as the update time for this target.
            if {[file isdirectory $target]} {
               set built($target) 0
            } else {
               set built($target) [file mtime $target]
            }
            return $built($target)
         }
         set emsg "Don't know how to make '$target'"
         set ei "$emsg\n"
         set ec NONE
         set prune "\n"
         incr numErrors
      }
      if {![info exists emsg]} {
         # Invoke the rule to build the target
         set built($target) pending  ;# safeguard against circularity
         if {![catch {$rule Invoke $target} msg]} {
            set built($target) $msg
            return $built($target)
         }
         foreach {ei ec} [list $errorInfo $errorCode] {}
         set emsg $msg
         set prune "\n    ('$class' instance method 'Invoke'"
      }
      set built($target) -1
      set index [string last $prune $ei]
      set ei [string trimright [string range $ei 0 $index]]
      error $emsg "$ei\n    (while building target '$target')" $ec
   }

   proc Define {definition} {
      # Note: This proc runs at global scope, even when called from
      #       a makerules.tcl running in a namespace.
      if {[string match {} [info script]]} {
         return -code error "'$class Define' must be in a makerules file"
      }

      set rulesFile [Oc_DirectPathname [file join [pwd] [info script]]]
      set filesRead($rulesFile) 1
      # Apply syntactic sugar
      set argstring ""
      foreach line [split $definition \n] {
         if {[info complete $argstring]} {
            append argstring " $line"
         } else {
            append argstring \n$line
         }
      }
      array set temp [uplevel list $argstring]
      # Filter out any unsupported options
      set args {}
      foreach opt [$class Options * New] {
         catch {lappend args $opt $temp($opt)}
      }
      catch {unset temp}
      array set temp $args
      # Supply additional constructor options
      if {[info exists temp(-ruleType)] \
             && [string match default $temp(-ruleType)]} {
         set directory [pwd]
      } else {
         set directory [file dirname $rulesFile]
      }
      array set temp [list -rulesFile $rulesFile -directory $directory \
                      -working_namespace [$class SelectNamespace $rulesFile]]
      if {[catch {[eval $class New _ [array get temp]] Initialize} msg]} {
         return -code error $msg
      }
   }

   # Search for rule which has the argument as a target.
   proc FindRule { target } {
      # If we already know a rule for this target, return it.
      if {[info exists ruleWhichMakes($target)]} {
         return $ruleWhichMakes($target)
      }
      # There could be a rule for this target in another makerules
      # file.  We search for the file 'makerules.tcl' starting in the
      # directory of the target upward until
      #  (1) We find a rule for the target
      #  (2) We find a directory w/o a makerules.tcl file
      #  (3) Sourcing makerules.tcl returns "continue"
      #  (4) We reach the root of the file system
      # The second option is meant to detect climbing past the start
      # of the oommf install tree, but would also fail if target lies
      # in the platform-specific directory that (usually) doesn't
      # contain a makerules.tcl file. To allow for this we start up
      # one level if the target directory doesn't have makerules.tcl.
      set dir [file dirname $target]
      if {![file exists [file join $dir makerules.tcl]]} {
         set dir [file dirname $dir]
      }
      while {![info exists ruleWhichMakes($target)]
             && [string compare $dir [file dirname $dir]]} {
         set rulesFile [file join $dir makerules.tcl]
         if {[catch {$class SourceRulesFile $rulesFile} msg]} {
            global errorInfo errorCode
            set index [string last "\n    ('$class' proc\
                        'SourceRulesFile'" $errorInfo]
            set ei [string trimright \
                       [string range $errorInfo 0 $index]]
            error $msg "$ei\n    (while finding rule for target\
                        '$target')" $errorCode
         }
         if {$msg == 0} {
            # Either no makerules.tcl file in this directory,
            # or else it returned "continue" when sourced.
            break
         }
         set dir [file dirname $dir]
      }
      # Check again to see if we now know a rule for the target.
      if {[info exists ruleWhichMakes($target)]} {
         return $ruleWhichMakes($target)
      }  else {
         return ""
      }
   }

   proc Offset {} {
      return $offset
   }

   proc SelectNamespace { target } {
      # See also class proc Shortname
      set leadpath [file dir [Oc_Main GetOOMMFRootDir]]
      set target [file dirname $target]
      if {[string match "${leadpath}*" $target]} {
         set target [string range $target [string length $leadpath]+1 end]
      }
      set target [string map {/ ::} [file join / $target]]
      return $target
   }

   proc RootOOMMFNamespace {} {
      return [$class SelectNamespace [file join [Oc_Main GetOOMMFRootDir] .]]
   }

   proc NamespaceDescendants { args } {
      # Returns a list of all descendants (children, children of
      # children, ...)  from a list of namespaces. If args is empty
      # then the OOMMF root namespace is inspected.
      if {[llength $args] == 0} {
         set args [list [$class RootOOMMFNamespace]]
      }
      set lst [list]
      foreach elt $args {
         lappend lst $elt
         set children [namespace children $elt]
         if {[llength $children]>0} {
            lappend lst {*}[$class NamespaceDescendants {*}$children]
         }
      }
      return $lst
   }

   proc SaveProc { name } {
      # Returns a Tcl script that can be used to recreate a proc in
      # another Tcl interpreter. Based on an example in the Tcl "info"
      # documentation.
      set params {}
      foreach var [info args $name] {
         if {[info default $name $var value]} {
            lappend params [list $var $value]
         } else {
            lappend params [list $var]
         }
      }
      return [list proc $name $params [info body $name]]
   }

   proc ChildInitScript {} {
      # Returns a Tcl script that can be exported to worker processes
      # to define procs and variables declared in makerules.tcl files.

      # Populate makerules.tcl-dependent namespaces
      set script {}
      foreach ns [$class NamespaceDescendants] {
         set wrk [list]
         foreach var [info vars ${ns}::*] {
            if {[info exists $var]} { ;# Variables might no be set
               lappend wrk [list set [namespace tail $var] [set $var]]
            } else {
               lappend wrk [list variable [namespace tail $var] [set $var]]
            }
         }
         foreach prc [info procs ${ns}::*] {
            set tmp [$class SaveProc $prc]
            lset tmp 1 [namespace tail $prc]
            lappend wrk $tmp
         }
         if {[llength $wrk]>0} {
            append script "namespace eval [list $ns] \{\n"
            append script [join $wrk "\n"]
            append script "\n\}\n"
         }
      }

      # Setup auto loading of procs from the Oc package w/o sourcing
      # pkg/oc/oc.tcl (which may create pkg/oc/tclIndex as a side
      # effect).
      set ocload [list]
      foreach {name val} [array get ::auto_index Oc_*] {
         if {![string match {* oc.tcl*} $val]} {
            lappend ocload $name $val ;# Exclude sourcing of oc.tcl
         }
      }
      append script "\narray set ::auto_index [list $ocload]\n"
      return [string trim $script]
   }


   proc SourceRulesFile {{rulesFile {}}} {
      if {[string match {} $rulesFile]} {
         set rulesFile makerules.tcl
      }
      set rulesFile [Oc_DirectPathname [file join [pwd] $rulesFile]]
      if {[info exists filesRead($rulesFile)]} {
         # We've already read the file.  No need to read it again.
         # Tell caller we've read the file.
         return $filesRead($rulesFile)
      }
      if {[file readable $rulesFile] && [file isfile $rulesFile]} {
         set filesRead($rulesFile) 1
         set wrkscript [subst {
            source [list $rulesFile]
            eval [list $defaultRules]
         }]
         set wns [$class SelectNamespace $rulesFile]
         set savedir [pwd]
         set code [catch {
            cd [file dirname $rulesFile]
            namespace eval $wns $wrkscript
         } msg]
         global errorInfo errorCode
         foreach {ei ec} [list $errorInfo $errorCode] {}
         cd $savedir
         if {$code == 1} { ;# error
            incr numErrors
         } elseif {$code == 4} {     ;# continue
            set filesRead($rulesFile) 0
            return 0
         } else {
            return 1
         }
      } else {
         set filesRead($rulesFile) 0
         return 0
      }
      set index [string last "\n    invoked from within" $ei]
      set ei [string trimright [string range $ei 0 $index]]
      error $msg "$ei\n    (reading makerules from file\
                '$rulesFile')" $ec
   }

   # A Tcl list of the targets of this MakeRule.
   # Targets must be strings which have the form of relative pathnames
   # which do not include the component '..'
   # This guarantees that each target refers to a (possibly mythical) file
   # in or below the directory containing the makerules file in which this
   # MakeRule is defined.
   #
   # PORTABILITY NOTE: The validation script will need revision if we ever
   # port to a platform on which it is some pathname component other than
   # '..' which means 'parent of this directory'.
   const public variable targets {
      if {[catch {set targets [lrange $targets 0 end]} msg]} {
         error "Invalid targets list: $targets
        Not a valid Tcl list: $msg"
      }
      foreach _ $targets {
         if {![string match relative [file pathtype $_]]} {
            error "Invalid target: $_
        All targets of a MakeRule must appear to be relative pathnames."
         }
         if {[lsearch -exact [file split $_] ..] >= 0} {
            error "Invalid target: $_
        MakeRule targets may not contain '..' as a pathname component."
         }
      }
   }

   # A Tcl list of the dependencies of the MakeRule, as given in the
   # MakeRule definition.
   # Each dependency should be a string in the form of a legal pathname,
   # which will be interpreted relative to the directory containing the
   # makerules file containing the definition of this MakeRule.
   const public variable dependencies = {} {
      foreach _ $dependencies {
         if {[catch {file split $_} __]} {
            error "Invalid dependency: $_
        $__"
         }
      }
   }
   # The dependencies of the rule expressed as direct, absolute pathnames
   private variable directDependencies
   method GetDirectDependencies {} {
      return $directDependencies
   }

   # The script to evaluate to update the targets when they are out of
   # date compared with the dependencies.
   const public variable script = {} {
      set script [string trim $script]
      if {![info complete $script]} {
         error "Value of -script option is not a complete Tcl script:
        $script"
      }
   }

   method HasNullScript {} { return [string match {} $script] }

   # The absolute pathname of the directory relative to which the
   # targets and dependencies are treated as filenames
   const public variable directory {
      # Possibly add other requirements here?
      if {[string match absolute [file pathtype $directory]] \
             && [file isdirectory $directory]} {
         # Ok!
      } else {
         error "Value of -directory option is not the absolute\
                pathname of a directory: $directory"
      }
   }

   # The file in which this rule is defined.
   const public variable rulesFile

   # The type of rule this is (plain or default)
   const public variable ruleType = plain {
      if {![string match plain $ruleType] \
             && ![string match default $ruleType]} {
         error "legal values: plain default"
      }
   }

   const public variable working_namespace

   Constructor {args} {
      eval $this Configure $args
   }

   method Initialize {} {
      foreach t $targets {
         set absTarget [file join $directory $t]
         if {![info exists defaultTarget]} {
            set defaultTarget $absTarget
         }
         if {[info exists ruleWhichMakes($absTarget)]} {
            if {[string match plain $ruleType]} {
               Oc_Log Log "Overriding previous rule for\
                            '$absTarget':\n\nOld rule\
                            was:\n[$ruleWhichMakes($absTarget) \
                            Description]\n\nNew rule is: [$this \
                            Description]\n" warning $class
               set ruleWhichMakes($absTarget) $this
            }
         } else {
            set ruleWhichMakes($absTarget) $this
         }
      }
      set directDependencies {}
      foreach d $dependencies {
         set dd [Oc_DirectPathname [file join $directory $d]]
         if {[lsearch -exact $directDependencies $dd]<0} {
            lappend directDependencies $dd  ;# No dups
            ## Another option would be to use 'lsort -unique',
            ## which would change the order. Would that be
            ## good or bad?
         }
      }
   }

   method Description {} {
      if {[string match default $ruleType]} {
         set msg "Default rule:\n"
      } else {
         set msg "In '$rulesFile':\n"
      }
      append msg "\tTargets: $targets\n\tDependencies:\
                $dependencies\n\tScript: $script"
      return $msg
   }

   method BuildDependencies {} {
      set latest 0
      foreach dep $directDependencies {
         if {[catch {
            set time [$class Build $dep]
         } msg]} {
            if {[info exists options(-k)]} {
               Oc_Log Log "Continuing after\
                            error:\n-----\n$msg\n-----" warning
               # Prevent building of dependent targets
               set time -1
            } else {
               global errorInfo errorCode
               foreach {ei ec} [list $errorInfo $errorCode] {}
               set emsg $msg
               set prune "\n    ('$class' proc 'Build'"
               break
            }
         }
         if {$latest >= 0} {
            if {$time < 0} {
               set latest -1
            } elseif {$time > $latest} {
               set latest $time
            }
         }
      }
      if {![info exists emsg]} {
         return $latest
      }
      set index [string last $prune $ei]
      set ei [string trimright [string range $ei 0 $index]]
      error $emsg $ei $ec
   }

   proc FlattenScript { script } {
      set outscript {}
      set cmd {}
      foreach line [split $script "\n"] {
         set line [string trim $line]
         if {"" eq $line} { continue }
         append cmd " $line"
         if {[info complete $cmd]} {
            append outscript "[string trim $cmd]\n"
            set cmd {}
         }
      }
      if {$cmd ne ""} {
         error "Bad script: \"$script\""
      }
      return [string trim $outscript]
   }

   method BareInvoke {} {
      global errorInfo errorCode
      # Be sure the script is evaluated with the current working
      # directory set to the directory containing the file which
      # contains the definition of this MakeRule
      set savedir [pwd]
      cd $directory
      set prior $numErrors
      if {![info exists options(-dry-run)]} {
         set code [catch {namespace eval $working_namespace $script} msg]
         foreach {ei ec} [list $errorInfo $errorCode] {}
      } else {
         if {[string length [set _ [string trim $script]]]>0} {
            puts stdout [$class FlattenScript $script]
         }
         set code 0
      }
      cd $savedir
      if {$code == 1} {
         # Don't double count errors when script invoke MakeRule
         # methods or procs
         if {$prior == $numErrors} {
            incr numErrors
         }
         error $msg $ei $ec
      }
   }

   method Invoke {target} {
      global errorInfo errorCode
      if {[info exists options(-d)]} {
         puts stdout "$pfx depends on:"
         append pfx "  "
      }
      if {[catch {$this BuildDependencies} latest]} {
         foreach {ei ec} [list $errorInfo $errorCode] {}
         set emsg $latest
         set prune "\n    ('$class' instance method\
                    'BuildDependencies'"
      }
      if {[info exists options(-d)]} {
         set pfx [string range $pfx 0 [expr {[string length $pfx] - 3}]]
      }
      # All dependencies up to date. 'latest' holds the most recent
      # modification time of any of them.  Check if that is more
      # recent than the modification time of our target.
      if {![info exists emsg] \
             && ((![file exists $target] && $latest >= 0) \
                    || ([file isfile $target] \
                           && [file mtime $target] < $latest))} {
         set code [catch {$this BareInvoke} msg]
         foreach {ei ec} [list $errorInfo $errorCode] {}
         if {$code == 1} {
            # Only halt processing on an error (return code = 1)
            # Ignore other return codes: ok, break, continue, return
            if {[info exists options(-i)]} {
               Oc_Log Log "Ignoring error:\n-----\n$msg\n-----" \
                  warning
            } else {
               set emsg $msg
               set prune "\n    (\"uplevel"
            }
         }
      }
      # The script was evaluated if necessary, report to caller the
      # build time of the target.
      if {![info exists emsg]} {
         if {![info exists options(-dry-run)] && [file isfile $target]} {
            return [file mtime $target]
         }
         if {[file isdirectory $target]} {
            return 0
         }
         return [expr {[clock seconds] + $offset}]
      }
      set index [string last $prune $ei]
      set ei [string trimright [string range $ei 0 $index]]
      error $emsg $ei $ec
   }

   ####################################################################
   ###             vvv   Parallel build code   vvv                  ###

   ###
   #
   # Comparison of serial and parallel pimake code:
   #
   #   The serial pimake code is a single-stage method, performing a
   # recursive depth-first traversal of the dependency graph, building and
   # updating timestamps as it goes.
   #
   #   The parallel pimake code runs in two stages. The first stage, run
   # serially, performs a recursive depth-first traversal of the dependency
   # graph, but doesn't build any targets at that time. Instead it just
   # saves the graph along with target timestamps. Any target that is older
   # than any of its dependencies, or descendants of those dependencies, is
   # marked for (re-)building. In the second stage all targets marked for
   # building are farmed out in parallel to separate processes, from the
   # bottom of the graph upwards, keeping track of dependency order to
   # ensure that dependents are built before parents.
   #
   #   One difference between the serial and parallel code is that the
   # latter makes all building decisions before any building commences.
   # A consequence is that for a dependency graph like
   #
   #         Targets:      foo   -->   foo.o   --> foo.cc
   #      Timestamps:   tomorrow --- yesterday --- today
   #
   # the older serial code will first rebuild foo.o, but then seeing that
   # foo is (still) newer than the just-built foo.o, will not rebuild the
   # executable foo.
   #
   #   OTOH, the parallel code will mark both foo and foo.o as out-of-date,
   # and rebuilds both.
   #
   #   Another difference is that in the serial build output from a
   # target's build process is written to stdout and stderr concurrent
   # with the build.  In the parallel build, the build output is
   # buffered and batched out only when a target build is
   # complete. This minimizes commingling of output from different
   # targets, but obscures build progress. This is most noticeable in
   # building the documentation, where the latex and HTML processing
   # can be quite lengthy.
   #
   #   Also of note: While serial pimake runs entirely as a single process
   # from the Tcl shell sourcing pimake.tcl, the parallel version launches
   # multiple independent Tcl shells (each running buildrule.tcl) to
   # perform the target builds. Separate shell processes are used instead
   # of child interpreters running inside the primary Tcl shell (e.g. via
   # the Thread package) because the build commands (compile, link, etc.)
   # may need to change the working directory, and the working directory
   # is a process-wide, not thread-specific, value. This causes trouble
   # if, e.g., two compiles running in parallel set their source-code
   # working directory to different paths. The parent process launches the
   # child processes using Tcl "open" command pipelines, and communicates
   # across the children's stdin/stdout channels.
   #
   ###

   private proc CombineDepTimes { t1 t2 } {
      # Implements the "Combining dependency timestamps" rule below.
      if {$t1<0 || $t2<0} { return -1 }
      return [expr {$t1>$t2 ? $t1 : $t2}]
   }

   proc RecurseDependencies { target upparent graphName } {
      # Note: target should be an absolute path
      #
      # Export graph is a dictionary where
      #      key : target
      #    value : A 6-element list ---
      #        (0) target timestamp
      #        (1) most recent dependency timestamp (as determined
      #            by the "inheritance" and "combining" timestamp
      #            rules below)
      #        (2) count of direct dependencies with newer timestamps
      #            than target (aka the unripe count)
      #        (3) list of dependencies
      #        (4) list of parents (i.e. targets w/ this target as dependency)
      #        (5) rule (i.e., MakeRule instance)
      #
      # All targets and dependencies have the form of an absolute path.
      #
      # Timestamp assignment rules:
      #     Existing regular files:     timestamp = file mtime
      #     Existing directories:       timestamp = 0
      #     Else if empty build script: timestamp = 0
      #     All other targets:          timestamp = -1
      #   Directories are treated specially because their mtime
      #   changes anytime a file is created or destroyed in them, but
      #   this shouldn't directly affect the timestamp properties of a
      #   companion file in the same directory. Non-existent files
      #   with no build script can't have a status change so they too
      #   are assigned the minimum time. The "all other targets"
      #   category are files that do not exist but are suppose to, or
      #   virtual targets such as "all" that have build scripts but
      #   never exist as entities on the filesystem.  The timestamp
      #   value -1 is treated as infinitely far in the past for build
      #   consideration (i.e., always build), but infinitely far in
      #   the future for dependency consideration (as in, this
      #   dependency will definitely be built, but at some point in
      #   the future).
      #
      #   Note the distinction between a target with no rule and a
      #   target with a rule and an empty build script. (Note that a
      #   rule Define w/o a -script argument is assigned an empty
      #   string as a default.)  A target with no rule should be an
      #   existing file (e.g., foo.cc), and is assigned a timestamp as
      #   above. A non-existing target with no rule will generate a
      #   "Don't know how to make <target>" error.
      #
      #   Targets with rules but w/ empty build scripts can still have
      #   and be dependencies, and serve to link other targets. For
      #   example, in distclean -> clean -> objclean, "clean" might
      #   have an empty build script.
      #
      # In the following discussion TT is the target timestamp, DT is
      # the dependency timestamp.
      #
      # Combining dependency timestamps:
      #   Given a collection of timestamps, Ti, the joint dependency time
      #   DT is -1 if Ti==-1 for any i, otherwise DT = max_i{Ti}.
      # This is implemented pairwise in the MakeRule proc CombineDepTimes
      #
      # Timestamp build rule:
      #   Given a target timestamp TT and the joint dependency time
      #   across all its dependencies, then build the target if
      #   (DT<0 || TT<DT).
      #
      # Dependency inheritance rule:
      #   Given a target timestamp TT and its joint dependency time DT,
      #   the inherited dependency time iDT presented to the parents of
      #   the target is (a) -1, if the target will be rebuilt, else (b)
      #   the result of combining TT and DT according to the timestamp
      #   combining rule above.

      upvar $graphName graph ;# Maybe not the best way to provide
      # common access to graph?
      set unripe_jobs [set ripe_jobs [list]]  ;# Return values
      if {[dict exists $graph $target]} {
         # Target already exists
         set data [dict get $graph $target]
         if {[llength $data]==0} {
            # Target is mid-construction. This indicates a cycle
            # in the dependency graph.
            error "Circular dependency: '$target'\
                (ultimately) depends on itself."
         }
         set timestamp [lindex $data 0]     ;# Return value (TT)
         set deptime   [lindex $data 1]     ;# Return value (DT)
         set unripe_count [lindex $data 2]  ;# Return value
         set parents   [lindex $data 4]
         if {[lsearch -exact $parents $upparent]>=0} {
            # This shouldn't happen provided each "directDependencies"
            # list is free of duplicates.
            error "Coding error? Entry $upparent already in parent\
                    list for target $target"
         }
         lappend parents $upparent
         lset data 4 $parents
         dict set graph $target $data
         # Note: unripe/ripe_job list entries should be handled when the
         # target node is created ("New target" stanza).
      } else {
         # New target
         # Extract rule info
         if {[catch {MakeRule FindRule $target} rule erropts]} {
            set errmsg "ERROR reading rules file in\
                         [file dirname $target]/: $rule"
            append errmsg "\n--- Stack ---\n[dict get $erropts -errorinfo]"
            append errmsg "\n--- Additional info: [dict get $erropts -errorcode]"
            error $errmsg
         }
         if {[string match {} $rule]} {
            # No rule for target. Throw an error if it is not a file.
            if {![file exists $target]} {
               error "No rule to build target $target"
            }
            set dependencies {}
            set has_build_script 0
         } else {
            # NB: Each MakeRule Define statement in a sourced
            #     makerules.tcl file creates a rule. If the Define does
            #     not specify -script then the default script is {}.
            set dependencies [$rule GetDirectDependencies]
            if {[string length [string trim [$rule Cget -script]]]==0} {
               set has_build_script 0  ;# Empty script
            } else {
               set has_build_script 1  ;# Non-empty script
            }
         }

         # Assign timestamp
         if {[file isfile $target]} {
            set timestamp [file mtime $target]
         } elseif {[file isdirectory $target]} {
            set timestamp 0
         } else {
            if {$has_build_script} {
               # Target does not exist (but could)
               set timestamp -1
            } else {
               # No build script --- linking target
               set timestamp 0
            }
         }

         dict set graph $target {} ;# Mark target as under construction.
         ## This is used to detect graph cycles.
         set deptime 0
         set unripe_count 0
         foreach dtarget $dependencies {
            lassign [$class RecurseDependencies $dtarget $target graph] \
               child_time child_deptime child_unripe_count \
               child_unripe_jobs child_ripe_jobs
            lappend unripe_jobs {*}$child_unripe_jobs
            lappend ripe_jobs   {*}$child_ripe_jobs
            if {$unripe_count>=0 && $child_unripe_count>=0} {
               if {$child_deptime<0 || $child_time<$child_deptime} {
                  incr unripe_count ;# Child is out-of-date
               }
            }
            if {$deptime >= 0} { ;# $deptime == -1 is already maximal
               if {[llength $child_ripe_jobs]>0} {
                  # If there are any ripe jobs, then there is at least
                  # one dependency that is out-of-date. When that
                  # dependency is rebuilt its timestamp will move into
                  # the future; denote this by setting deptime to
                  # -1. (This is true too if there are any unripe jobs,
                  # but in that case DT==-1 will be set at a lower level
                  # and propagate upward regardless.)
                  set deptime -1
               } else {
                  set checktime \
                     [MakeRule CombineDepTimes $child_time $child_deptime]
                  set deptime \
                     [MakeRule CombineDepTimes $deptime $checktime]
               }
            }
         }

         dict set graph $target [list $timestamp $deptime $unripe_count \
                                    $dependencies $upparent $rule]
         # Job assignment.
         if {$unripe_count>=0} {
            if {$deptime<0 || $timestamp<$deptime} {
               # Target is out-of-date
               if {$unripe_count>0} {
                  lappend unripe_jobs $target
               } else {
                  lappend ripe_jobs $target
               }
            }
         }
      }
      return [list $timestamp $deptime $unripe_count $unripe_jobs $ripe_jobs]
   }
   proc ListAllParents { graph target } {
      set parents [lindex [dict get $graph $target] 4]
      foreach parent $parents {
         foreach upparent [$class ListAllParents $graph $parent] {
            if {[lsearch -exact $parents $upparent]<0} {
               lappend parents $upparent
            }
         }
      }
      return $parents
   }
   proc ConstructBuildGraph { graph job_list } {
      # Construct the build graph bottom-up from the job_list.
      set buildgraph [dict create]
      # Select up-targets from job_list
      foreach job $job_list {
         dict set buildgraph $job [dict get $graph $job]
         set lap [$class ListAllParents $graph $job]
         foreach parent $lap {
            dict set buildgraph $parent [dict get $graph $parent]
         }
      }
      # Prune unaccessed dependencies
      dict for {key val} $buildgraph {
         set deps [lindex $val 3]
         set pruned_deps {}
         foreach dep $deps {
            if {[dict exists $buildgraph $dep]} {
               lappend pruned_deps $dep
            }
         }
         lset val 3 $pruned_deps
         dict set buildgraph $key $val
      }
      return $buildgraph
   }
   proc TimestampPrintSuffix { target graph } {
      if {$options(-verbose)<2} { return "" }
      lassign [dict get $graph $target] TT DT UC DL PL RULE
      if {$DT<0} {
         set cmp ":"
      } else {
         set cmp "="
         if {$TT<$DT} {set cmp "<"} elseif {$TT>$DT} {set cmp ">"}
      }
      set suffix " ($TT $cmp $DT $UC)"
      if {$options(-verbose)>4} {
         set scriptstr [string trim [$RULE Cget -script]]
         if {[string match {} $scriptstr]} {
            append suffix " --- No script"
         } else {
            append suffix "\n--- Script ---\n$scriptstr\n---"
         }
      }
      return $suffix
   }
   proc Shortname { target } {
      # See also class proc SelectNamespace
      set leadpath [file dir [Oc_Main GetOOMMFRootDir]]
      if {[string match "${leadpath}*" $target]} {
         set target [string range $target [string length $leadpath]+1 end]
      }
      return $target
   }
   proc PrintDependencies { prefix target graph fancygraphics } {
      if {![dict exist $graph $target]} {
         return ;# No dependencies
      }
      set dependencies [lindex [dict get $graph $target] 3]
      if {$fancygraphics} {
         # Use unicode box drawing characters
         set tee \u251C ; set bar \u2500 ; set pip \u2502 ; set cor \u2514
      } else {
         # Use plain ASCII characters
         set tee "+"    ; set bar "-"    ; set pip "|"    ; set cor "+"
      }
      set depcnt [llength $dependencies]
      for {set i 0} {$i<$depcnt} {incr i} {
         set dep [lindex $dependencies $i]
         set shortname [$class Shortname $dep]
         set sfx [$class TimestampPrintSuffix $dep $graph]
         if {$i<$depcnt-1} {
            puts stdout "${prefix}$tee$bar$bar$shortname$sfx"
            $class PrintDependencies "${prefix}$pip   " \
               $dep $graph $fancygraphics
         } else {
            puts stdout "${prefix}$cor$bar$bar$shortname$sfx"
            $class PrintDependencies "${prefix}    " \
               $dep $graph $fancygraphics
         }
      }
   }
   proc ConstructDependencyGraph { {target {}} } {
      # Note: Single target only
      if {[string match {} $target]} {
         set target $defaultTarget
      }
      set target [Oc_DirectPathname $target]
      set graph [dict create]
      lassign [$class RecurseDependencies $target {} graph] \
         timestamp deptime unripe_count unripe_jobs ripe_jobs

      # If stdout encoding is utf-8, then use unicode box drawing
      # characters to represent graph connections. Otherwise fall
      # back to plain ASCII.
      set fancygraphics 0    ; set bar "-"
      if {[fconfigure stdout -encoding] eq "utf-8"} {
         set fancygraphics 1 ; set bar \u2500
      }
      set titlenote ""
      if {$options(-verbose)>2} {
         set titlenote "(out of date on < or :) "
      }
      if {[info exists options(-depgraph)]} {
         puts stdout "$bar$bar$bar Dependency Graph $titlenote$bar$bar$bar"
         set sfx [$class TimestampPrintSuffix $target $graph]
         puts stdout "Target: [$class Shortname $target]$sfx"
         $class PrintDependencies " " $target $graph $fancygraphics
         puts stdout [string repeat $bar 24]
      }
      if {[info exists options(-buildgraph)]} {
         puts stdout "$bar$bar$bar Build Graph $titlenote$bar$bar$bar"
         set sfx [$class TimestampPrintSuffix $target $graph]
         puts stdout "Target: [$class Shortname $target]$sfx"
         set buildgraph [$class ConstructBuildGraph $graph \
                            [concat $unripe_jobs $ripe_jobs]]
         $class PrintDependencies " " $target $buildgraph $fancygraphics
         puts stdout [string repeat $bar 20]
      }
      return [list $graph $unripe_jobs $ripe_jobs]
   }

   # Note: It might be a good idea to set up a separate class to manage
   #       the separate build processes, communication, and buffers.

   # Buffer used to hold build command response communications. One
   # buffer for each channel, indexed by the channel id.
   private common chanbuf

   # Semaphore used to signal command completion
   private common rule_done

   # Code for writing to and reading from job support process.  These
   # are not symmetric with each other, but rather complementary with
   # the ReadMessage and WriteMessage* routines in buildrule.tcl.
   proc WriteCmd { chan cmd } {
      set msgid [expr {int(rand()*pow(2,53))}]
      puts $chan \
         "MESSAGE START ID $msgid\n$cmd\nID $msgid MESSAGE END"
      flush $chan
   }
   # Return from ReadMessage is a two element list. The first element a
   # result code and the second is the import buffer appended with input
   # from chan. The result codes are
   #    0 => Success. In this case the second element is a two element
   #         list consisting of the cmd return code and the result.
   #    1 => Partial message; pass buffer back on next readable event
   #    2 => EOF on channel
   #    3 => Mangled message
   proc ReadCmdResult { chan buffer } {
      while {[gets $chan line]>=0} {
         # Read one line at a time, stopping when either data runs out
         # or else the end-of-message marker is detected. This allows
         # separation of two consecutive messages, which can occur if
         # the second is a timeout message from the child.
         append buffer "$line\n"
         if {[regexp { MESSAGE END$} $line]} {
            break
         }
      }
      if {[catch {eof $chan} status] || $status} {
         catch {close $chan}
         return [list 2 {}]
      }
      set endpt [string first "MESSAGE END\n" $buffer]
      if {$endpt<0} {
         # Partial message. Client should call back when channel is
         # readable again.
         return [list 1 $buffer]
      }
      set buf [string trim $buffer]
      if {![regexp {^MESSAGE START ID (\S+)\n(.*)$} $buf _ msgid tail] \
             || ![regexp "^(.*)ID $msgid RESULT (\\S+) MESSAGE END$" \
                    $tail _ msg msgcode]} {
         return [list 3 $buffer]
      }
      return [list 0 [list $msgcode [string trim $msg]]]
   }

   # Wrapper for ReadCmdResult. Result is sent through the rule_done
   # semaphore, as a four element list:
   #   thread_index communication_errorcode rule_errorcode rule_output
   # The communication_errorcode field will be either 0 (success), 2
   # (EOF), or 3 (mangled message).
   proc ReadCmdResultHandler { index chan } {
      lassign [$class ReadCmdResult $chan [dict get $chanbuf $chan]] code buf
      if {$code == 0} { ;# Message complete
         dict set chanbuf $chan {}  ;# Clear buffer for next message
         lassign $buf cmdcode result
         lappend rule_done \
            [list $index 0 $cmdcode $result] ;# Trigger semaphore
         return
      }
      if {$code == 1} { ;# Incomplete message
         dict set chanbuf $chan $buf
         return
      }
      # Channel communication error
      lappend rule_done \
         [list $index $code 0 {*}$buf] ;# Trigger semaphore
      return
   }
   proc ReadChildReadyHandler { index chan } {
      while {[gets $chan line]>=0} {
         # Read and ignore all input on chan until "Child Ready" is
         # received. This allows stdout/stderr chatter on the pipe
         # arising from the child initialization to be ignores.
         if {$options(-verbose)>4} {
            puts [format "CHILD %2d INIT: %s" $index $line]
         }
         if {[string compare "Child Ready" $line]==0} {
            fileevent $chan readable \
               [list $class ReadCmdResultHandler $index $chan]
            return [$class ReadCmdResultHandler $index $chan]
         }
      }
      if {[catch {eof $chan} status] || $status} {
         catch {close $chan} msg
         if {$options(-verbose)>4} {
            puts [format "CHILD %2d INIT: Premature eof: %s" $index $msg]
         }
         lappend rule_done [list $index 2 0 $msg] ;# Trigger semaphore
         return
      }
   }
   proc OpenCmdChan { index export_script } {
      set TCLSH [info nameofexecutable]
      set cmdfile [list]
      lappend cmdfile | $TCLSH
      lappend cmdfile [file join [file dirname [info script]] buildrule.tcl]
      lappend cmdfile $index
      lappend cmdfile {*}[array get options]
      if {[string length $export_script]>0} {
         lappend cmdfile -initscript $export_script
      }
      lappend cmdfile "2>@1"
      set chan [open $cmdfile r+]
      fconfigure $chan -blocking 0 -buffering full -buffersize 1000000
      fileevent $chan readable "$class ReadChildReadyHandler $index $chan"
      return $chan
   }

   proc CloseCmdChan { chan } {
      # NB: If child buildrule.tcl processes are left hanging in some
      #     irregular exit cases, one can experiment with tying each
      #     child channel to an Oc_EventHandler firing on Oc_Main
      #     Shutdown events.
      set pid "?"
      if {[catch {
         set pid [pid $chan]
         #fconfigure $chan -blocking 1
         close $chan
      } errmsg]} {
         puts $errchan "Command channel (pid=$pid) close error: $errmsg"
      }
   }

   method SubmitCmd { job index chan } {
      # AFAICT, all rule targets are singular items, so
      # $job and $targets should be the same.
      set build_dir [list $directory]
      if {[info exists options(-dry-run)]} {
         set flat_script [$class FlattenScript $script]
         if {[string length $flat_script]==0} {
            set cmd {}
         } else {
            set cmd [list puts stdout $flat_script]
         }
      } else {
         set cmd [list namespace eval $working_namespace $script]
      }
      $class WriteCmd $chan [subst -nocommands {
         set errmsg {}
         set errinfo {}
         set errcode 0
         set savedir [pwd]
         if {[catch {
            cd $build_dir
            $cmd
         } msg erropts]} {
            set errmsg \$msg
            catch { set errinfo [dict get \$erropts -errorinfo] }
            set errcode 2
         }
         cd \$savedir
         list $job $index \$errcode \$errmsg \$errinfo
      }]
   }


   # Parallel-build background error handler
   proc PB_BackgroundError { errmsg opts } {
      if {[dict exists $opts -errorinfo]} {
         puts stderr "FATAL BACKGROUND ERROR: [dict get $opts -errorinfo]"
      } else {
         puts stderr "FATAL BACKGROUND ERROR: $errmsg"
      }
      exit 99
   }

   proc ParallelBuild { {target {}} } {
      interp bgerror {} [list $class PB_BackgroundError]
      set max_thread_count $options(-parallel)
      if {$max_thread_count<1} {
         error "Bad -parallel value: $options(-parallel)"
      }
      set verbose $options(-verbose)

      # Build dependency graph
      # Note: The client code in pimake.tcl should run
      #    MakeRule SourceRulesFile
      # prior to calling ParallelBuild.
      lassign [MakeRule ConstructDependencyGraph $target] \
         graph unripe_jobs ripe_jobs

      set target_key [lindex [dict keys $graph] 0]
      set target_time [lindex [dict get $graph $target_key] 0]
      if {$target_time < 1} {
         # Special case: Directory or non-existing target
         set target_time [clock milliseconds]
      }

      if {$verbose>1} {
         puts stdout "--- Jobs waiting on dependency build (unripe):"
         foreach j $unripe_jobs {
            puts stdout " [regsub {^.*oommf/} $j oommf/]"
         }
         puts stdout "--- Jobs ready to run (ripe):"
         foreach j $ripe_jobs {
            puts stdout " [regsub {^.*oommf/} $j oommf/]"
         }
         puts stdout "---"
      }

      if {[info exists options(-depgraph)] \
             || [info exists options(-buildgraph)]} {
         exit
      }

      if {$verbose>4} {
         puts "Makerule export script--\n[$class ChildInitScript]\n---"
      }

      # Initialize child process support variables
      set export_script [$class ChildInitScript]
      set rule_done [list]
      set chanbuf [dict create]  ;# Maps channels to buffer space
      set build_threads {}
      set max_job_time [list 0 {}]
      set max_child_count 0
      ## Each entry in the build_threads list is a six element list:
      ##  (0) Assigned channel; set to empty string on eof
      ##  (1) Channel state: 0=>inactive, 1=>active, 2=>error,
      ##      -1=>assigned job not received by child
      ##  (2) Job (i.e., target) or empty string if none assigned
      ##  (3) dispatch time
      ##  (4) job completion time
      ##  (5) Run output from job (may include error message)

      # Loop through jobs
      set timeout 600000  ;# 10 minutes; adjust if any jobs take
                          ## longer than this.
      if {[info exists options(-i)]} {
         set continue_on_error 2
      } elseif {[info exists options(-k)]} {
         set continue_on_error 1
      } else {
         set continue_on_error 0
      }
      set job_errors 0
      set bad_jobs {}
      set active_count 0
      set timeout_error 0
      while {[llength $ripe_jobs]>0 || $active_count>0} {
         # START outer jobs loop
         while {$active_count<$max_thread_count && \
          [llength $ripe_jobs]>0 && \
          ($job_errors==0 || $continue_on_error)} { ;# START job submit loop

            # Pop first job
            set ripe_jobs [lassign $ripe_jobs job]
            lassign [dict get $graph $job] timestamp dep_timestamp \
               unripe_count dependencies parents rule

            # Check script
            set rule [$class FindRule $job]
            if {[$rule HasNullScript]} {
               # Shortcut handling of empty scripts.
               # Note: These count as "active jobs" wrt active_count,
               #       which is used for loop control, but don't enter
               #       the build_threads list. These jobs go straight
               #       to the rule_done semaphore w/o passing through
               #       the event loop, so they should (must!?) be
               #       completely cleared out during the rule_done
               #       processing loop.
               lappend rule_done [list -1 $job [clock milliseconds]]
               if {$verbose>2} {
                  puts "Rule submit: Shortcut $job"
               }
            } else {
               # Find free thread
               set free_index \
                  [lsearch -exact -index 1 -integer $build_threads 0]
               if {$free_index<0 || \
                   [string match {} [lindex $build_threads $free_index 0]]} {
                  # Open child process channel
                  if {$free_index<0} {
                     # Append new channel slot
                     if {[dict size $chanbuf] >= $max_thread_count} {
                        puts "--- THREAD ERROR ---"
                        foreach elt $build_threads {
                           puts $elt
                        }
                        puts "--------------------"
                        error "Programming error: No free threads"
                     }
                     set free_index [llength $build_threads]
                     lappend build_threads {} ;# Dummy elt for lset below
                     set chaninfo "new channel"
                  } else {
                     # Otherwise we are reusing an existing slot
                     set chaninfo "channel restart"
                  }
                  set chan [$class OpenCmdChan $free_index $export_script]
                  lset build_threads $free_index [list $chan 0 {} {} {} {}]
                  dict set chanbuf $chan {}
                  if {[dict size $chanbuf] > $max_child_count} {
                     set max_child_count [dict size $chanbuf]
                  }
                  if {$verbose>2} {
                     puts stdout "Opening command channel $free_index/$chan:\
                          Child process [pid $chan] started ($chaninfo)."
                  }
               }
               set chan [lindex $build_threads $free_index 0]
               lset build_threads $free_index [list $chan 1 $job \
                                               [clock milliseconds] {} {}]
               if {$verbose>2} {
                  puts "Rule submit: channel $free_index:\
                        [lindex $build_threads $free_index]"
               }

               if {[catch {$rule SubmitCmd $job $free_index $chan} wrtmsg]} {
                  # Error writing cmd request to submit channel. This
                  # can happen if a command channel shuts down due to
                  # inactivity while the mainline is inside the submit
                  # loop. There should be an EOF on the channel waiting
                  # in the event loop, so put the job back on the to-do
                  # list and break out of the job submit loop to allow
                  # event loop processing.
                  if {$verbose>2} {
                     puts "--> Submit fail on channel\
                               $free_index: $wrtmsg (retry)"
                  }
                  catch {fileevent $chan readable {}}
                  catch {close $chan}
                  dict unset chanbuf $chan  ;# This shouldn't fail
                  unset chan
                  lset build_threads $free_index [list {} 0 {} {} {} {}]
                  lappend ripe_jobs $job
                  continue
               }
            }
            incr active_count
         } ;# END job submit loop

         if {$active_count < 1} {
            break ;# Exit outer jobs loop
         }

         # Wait for some rule to complete. Class common rule_done is
         # used as a semaphore when a task completes. The contents of
         # rule_done is a list of lists. Normal (non-shortcut) jobs
         # should be a four element list:
         #
         #   thread_index communication_errorcode rule_errorcode rule_output
         #
         # The communication_errorcode field will be either 0 (success),
         # 2 (EOF), or 3 (mangled message).
         #  In principle, a communication error might be overcome by
         # starting a new channel and re-running the rule, but the error
         # source is most likely an uncatchable error processing the
         # rule that leads to an abort of the Tcl shell and closing of
         # the command channel. In this case case a re-run would simply
         # repeat the failure, so the code treats this the same as a
         # rule error.
         #  Shortcut jobs (i.e., ones with empty build scripts) append
         # a three element list on the rule_done FIFO queue:
         #
         #   -1 job dispatch_time
         #
         # In this case the append occurs directly, w/o traversing the
         # event loop.
         if {[llength $rule_done]==0} {
            # Wait for something to process
            set timeout_id [after $timeout \
                               "set [$class GlobalName rule_done] TIMEOUT"]
            vwait [$class GlobalName rule_done]   ;# Semaphore
            after cancel $timeout_id
            if {[string compare $rule_done TIMEOUT]==0} {
               puts $errchan [format "ERROR: Timeout (%g s) waiting for\
                       command channel response" [expr {$timeout/1000.}]]
               set timeout_error 1
               break  ;# Exit outer jobs loop
            }
         }
         while {[llength $rule_done]>0} { ;# START rule_done FIFO processing
            set completion_time [clock milliseconds]
            set rule_done [lassign $rule_done results] ;# Pop rule_done FIFO
            if {$verbose>2} {
               puts "Rule report loop, active_count=$active_count"
               puts "Rule report: [list $results]"
            }
            set results [lassign $results done_index] ;# Extract thread index
            if {$done_index == -1} {
               # Shortcut, empty build script
               lassign $results done_job dispatch_time
               set done_status 1 ;# Safety
               set comm_code [set cmd_code 0]
               set cmd_response {}
               incr active_count -1
            } else {
               # Usual case: non-empty build script
               lassign $results comm_code cmd_code cmd_response
               lset build_threads $done_index 4 $completion_time
               lassign [lindex $build_threads $done_index] \
                  chan done_status done_job \
                  dispatch_time completion_time done_job_response
               # The completion_time slot in build_threads is reserved
               # for future use. In particular, it could be set by the
               # command channel handler and then extracted here.

               if {[regexp {^CHILD TIMEOUT ([0-9]+)$} $cmd_response \
                       _ child_pid]} {
                  # Response is not from a task assignment, but rather
                  # a child has timed out and is in the process of
                  # shutting down. A subsequent EOF should appear on
                  # the channel, but we shortcircuit that by removing
                  # the event handler and closing the channel. In this
                  # setting any assigned jobs were not processed by
                  # the child, so move them back to the ripe list.
                  catch {fileevent $chan readable {}}
                  catch {close $chan}
                  if {[catch {dict unset chanbuf $chan}]} {
                     puts stderr "Unexpected chanbuf\
                        ($chan/pid=$child_pid) error; ignoring"
                  }
                  unset chan
                  lset build_threads $done_index [list {} 0 {} {} {} {}]
                  if {[string match {} $done_job]} { ;# Idle channel
                     if {$verbose>2} {
                        puts stdout \
                           "--> Command channel $done_index: closed (idle)"
                     }
                  } else { ;# Reset job
                     lappend ripe_jobs $done_job
                     incr active_count -1
                     if {$verbose>2} {
                        puts stdout "--> Command channel $done_index:\
                                  closed; resubmitting $done_job"
                     }
                  }
                  continue
               } elseif {$done_status != 0} {
                  # Only decrement active_count if rule_done was
                  # triggered by job completion (with or without
                  # error) from an active thread. In particular,
                  # ignore timeouts from inactive threads.
                  incr active_count -1
               }

               # Reset thread state
               lset build_threads $done_index [list $chan 0 {} {} {} {}]
            }

            if {![string match {} $dispatch_time]} {
               # dispatch_time may be empty on error
               set job_run_time [expr {$completion_time - $dispatch_time}]
               if {$job_run_time > [lindex $max_job_time 0]} {
                  set max_job_time [list $job_run_time $done_job]
               }
            }

            if {$comm_code != 0 || $cmd_code != 0} {
               # Failure
               if {$comm_code != 0} {
                  # Communication error; reset channel
                  catch {fileevent $chan readable {}}
                  catch {close $chan}
                  catch {dict unset chanbuf $chan}
                  unset chan
                  lset build_threads $done_index [list {} 0 {} {} {} {}]
                  if {[string match {} $done_job]} {
                     # Command channel was idle.
                     if {$verbose>2} {
                        puts stdout \
                         "--> Command channel $done_index: closed (idle eof)"
                     }
                     continue
                  } else {
                     # Error occurred on channel with an assigned job.
                     # On controlled timeout a "CHILD TIMEOUT" message
                     # should precede an EOF (comm_code == 2). The
                     # "CHILD TIMEOUT" handling will disconnect the
                     # fileevent handler on the channel, so the only
                     # way to get here is if the eof was already
                     # pulled off the event loop before this "START
                     # rule_done" loop was entered. In that case the
                     # build_threads job will be set empty by the
                     # "CHILD TIMEOUT" handling earlier in the same
                     # "START rule_done" loop bundle, so the preceding
                     # if branch will operate. So best guess is that
                     # the root cause of the comm error was the build
                     # script and a resubmit would suffer the same
                     # fate (i.e., fatal error).
                     set errmsg "Error on cmd channel $done_index:\
                      (comm error $comm_code): $cmd_response"
                     append errmsg " : Target $done_job not completed"
                  }
               } else {
                  # Rule error
                  set errmsg \
                     "Rule error: $cmd_response\n  on target: $done_job"
               }
               incr job_errors
               lappend bad_jobs $done_job

               # If an error occurred and $continue_on_error is false, then
               # continue processing active jobs and stop.  If an error
               # occurred and $continue_on_error is 1 (corresponding to
               # options(-k)), then don't adjust parent unripe counts but
               # otherwise continue.  If an error occurred and
               # $continue_on_error is 2 (corresponding to options(-i)),
               # then update parent unripe counts and continue.
               if {$continue_on_error} {
                  # CoE == 1 (continue after errors)
                  #  or == 2 (ignore errors)
                  puts $errchan "----- CONTINUING AFTER\
                                 ERROR:\n      $errmsg\n-----"
               } else {
                  # CoE == 0 => Don't continue
                  puts $errchan "----- TERMINAL ERROR: \n$errmsg\n-----"
               }
               if {$continue_on_error<2} {
                  continue ;# Don't update
               }
            } else {
               # Rule success
               set cmd_response [string trim $cmd_response]
               if {[string length $cmd_response]>0} {
                  puts stdout $cmd_response
               }
            }

            if {$verbose>2} {
               puts "Job  complete channel $done_index: $done_job"
            }
            if {$done_job eq $target_key} {
               set target_time $completion_time
            }

            if {$job_errors == 0 || $continue_on_error} {
               # Update task list
               lassign [dict get $graph $done_job] timestamp dep_timestamp \
                  unripe_count dependencies parents rule
               foreach p $parents {
                  # Decrement unripe count in each parent
                  set gp [dict get $graph $p]
                  set gpcnt [lindex $gp 2]
                  incr gpcnt -1
                  lset gp 2 $gpcnt
                  dict set graph $p $gp
                  if {$gpcnt==0} {
                     # Move parent from unripe to ripe list.
                     lappend ripe_jobs $p
                     set index [lsearch -exact $unripe_jobs $p]
                     if {$index<0} {
                        error "Programming error: Job $p missing from\
                         unripe_jobs list"
                     }
                     set unripe_jobs [lreplace $unripe_jobs $index $index]
                  }
               }
            }

         } ;# END rule_done queue processing
         if {$verbose>2} {
            puts stdout [format "Bottom of outer jobs loop:%3d ripe jobs,\
                   %2d active threads" [llength $ripe_jobs] $active_count]
         }
      } ;# END outer jobs loop
      set unfinished_jobs [concat $unripe_jobs $ripe_jobs]
      if {($continue_on_error || [llength $bad_jobs]==0 \
              || $verbose>1) && [llength $unfinished_jobs]>0} {
         set errmsg "--- WARNING: Unrun jobs -->\n"
         append errmsg [join $unfinished_jobs "\n"]
         append errmsg "\n<--- Unrun jobs"
         puts $errchan $errmsg
      }
      if {[llength $bad_jobs]>0} {
         set errmsg "--- ERROR building target(s):\n"
         append errmsg [join $bad_jobs "\n"]
         append errmsg "\n---"
         puts $errchan $errmsg
      }

      if {$verbose>4} {
         puts "Build closing status:"
         puts "--- build_threads ---"
         set index 0
         foreach elt $build_threads {
            puts [format "%2d : %s" $index $elt]
            incr index
         }
         puts "---------------------"
         puts "------ chanbuf ------"
         foreach key [dict keys $chanbuf] {
            puts [format "%12s : %s" $key [dict get $chanbuf $key]]
         }
         puts "---------------------"
      }

      # Close command channels
      set chanlist {} ; set pidlist {}
      set chanbufkeys [dict keys $chanbuf]
      foreach chan $chanbufkeys {
         if {![catch {pid $chan} pid]} {
            lappend chanlist $chan
            lappend pidlist $pid
         } elseif {$verbose>2} {
            puts $errchan "Warning: $pid"
         }
      }
      if {$verbose>2} {
         puts stdout \
            "Closing [llength $pidlist] channels; pids: [join $pidlist]"
      }
      foreach chan $chanlist {
         $class CloseCmdChan $chan
      }

      if {$verbose>1} {
         lassign $max_job_time time job
         if {$time > 1} {
            set job [$class Shortname $job]
            set time [format %.2f [expr {$time/1000.}]]
            puts stdout "Longest running job: $job ($time seconds)"
         }
         puts stdout "Maximum parallel job count: $max_child_count"
      }

      set numErrors $job_errors
      if {$numErrors == 0 || $continue_on_error>1} {
         return [expr {int(round($target_time))}]
      }
      error "Error building target $target_key"
   }
   ###
   ###             ^^^   Parallel build code   ^^^                  ###
   ####################################################################
}
