# FILE: platform.tcl
#
# The Platform class -- no instances, just a way to provide a "major minor"
# style of Tcl command.

Oc_Class Platform {

    public common cflags {}
    public common lflags {}
    public common noexedelete 0

    const common configuration
    const common platformName
    const common oommfpkg

    array common linkopts
    array common compopts
    array common mlopts

    ClassConstructor {
        set configuration [Oc_Config RunPlatform]
        set platformName [$configuration GetValue platform_name]
        set oommfpkg [file join [Oc_Main GetOOMMFRootDir] pkg]
        set linkopts(obj) {Platform Objects}
        set linkopts(out) {Platform Executables}
        set linkopts(lib) {Platform LinkerLibs}
        set compopts(out) {Platform Objects}
        set compopts(src) {Platform AbsoluteSrc}
        set mlopts(obj) {Platform Objects}
        set mlopts(out) {Platform PlatformLibrary}
    }

    #    proc EvalSafeList {l} {
    #        eval list \{[join $l "\} \{"]\}
    #    }
    #
    #    The attempt above may be sufficient for our purposes, but it can't
    #    handle unmatched open brace as an element.  The replacement is safe.
    proc EvalSafeList {l} {
        set r {}
        foreach e $l {
            lappend r $e
        }
        return $r
    }

    proc Name {} {
        return $platformName
    }

    proc Specific {files} {
        set ret {}
        foreach f $files {
            lappend ret [file join $platformName $f]
        }
        return $ret
    }

    proc Executables {stemlist} {
        set ret {}
        set script [$configuration GetValue script_filename_executable]
        if {[string length $script]} {
            foreach stem $stemlist {
                lappend ret [file join $platformName [eval $script $stem]]
            }
        }
        return $ret
    }

    proc Intermediate {stemlist} {
        set ret {}
        if {[catch {
            set scripts [$configuration GetValue script_filename_intermediate]
        }]} {
            return $ret
        }
        foreach script $scripts {
            if {[string length $script]} {
                foreach stem $stemlist {
                    lappend ret [file join $platformName [eval $script $stem]]
                }
            }
        }
        return $ret
    }

    proc Objects {stemlist} {
        set ret {}
        set script [$configuration GetValue script_filename_object]
        if {[string length $script]} {
            foreach stem $stemlist {
                lappend ret [file join $platformName [eval $script $stem]]
            }
        }
        return $ret
    }

    # The static/shared platform scripts convert a library stem to
    # the file name form used on the platform. For example, on Linux
    #    Static libraries:  oc -> liboc.a
    #    Shared libraries: tcl -> libtcl.so
    #
    # The static stem script is required on all platforms; the shared
    # script is optional. On Linux, a stem starting with a colon is
    # assumed to be a pre-formatted, full library name. For example,
    #
    #    :libsundials_cvode.so.6 -> libsundials_cvode.so.6
    #
    # This can be used to force load a specific library version.

    # The PlatformLibrary class proc class is used to construct the file
    # name for an OOMMF pkg library from a pkg stem.  It runs the
    # platform static library script to form the full library name and
    # then prepends the platform name, e.g., on 64-bit Linux
    #
    #    oc -> linux-x86_64/liboc.a
    #
    # The returned value is a relative path taken wrt the package
    # root. This allows the value to be used as a target in
    # makerules.tcl files.
    proc PlatformLibrary {stem} {
       set ret {}
       set script [$configuration GetValue script_filename_static_library]
       if {[string length $script]} {
          lappend ret [file join $platformName [eval $script $stem]]
       }
       return $ret
    }

    proc FindOommfLibraries { stemlist } {
       # Returns a list of resolved filenames for OOMMF pkg libraries.
       # Example stemlist: {vf xp nb oc}.
       # NB: At present all OOMMF pkg libraries are static libraries.
       set ret {}
       set script [$configuration GetValue script_filename_static_library]
       if {[string length $script]} {
          foreach stem $stemlist {
             set pkgdir [file join $oommfpkg $stem]
             if {[file isdirectory $pkgdir]} {
                lappend ret [file join $pkgdir $platformName \
                                 [eval $script $stem]]
             }
          }
       }
       return $ret
    }

    proc FindLibraries { stemlist } {
       # Returns a list of resolved filenames matching import stemlist.
       # This proc first checks for OOMMF pkg libraries, then system
       # shared libraries, and finally system static libraries. Here
       # "system" means libraries provided outside of OOMMF.
       set ret {}
       foreach stem $stemlist {
          set chk [$class FindOommfLibraries [list $stem]]
          if {[llength $chk]} {
             lappend ret $chk;
             continue
          }

          # Otherwise, stem is not a recognized pkg stem. See if it is
          # an external library in one of the extra library directories,
          # in which case include the relevant directory on the link
          # command line. Note: We only need to check the configuration
          # value program_linker_extra_lib_dirs since libraries on the
          # system library search path will be automatically picked up
          # by the linker.
          set libdirs [list]
          if {![catch {$configuration GetValue \
                          program_linker_extra_lib_dirs} chk]} {
             lappend libdirs {*}$chk
          }
          if {[llength $libdirs]<1} {
             continue  ;# No libs to search
          }
          set checkscripts [list]
          if {![catch {$configuration GetValue \
                          program_linker_extra_lib_scripts} _]} {
             foreach script $_ {
                lappend checkscripts [list $_ glob]
                # "Extra" scripts return a list of glob patterns
             }
          }
          if {![catch {$configuration GetValue \
                          script_filename_shared_library} _]} {
             lappend checkscripts [list $_ exact]
          }
          if {![catch {$configuration GetValue \
                          script_filename_shared_library_glob} _]} {
             lappend checkscripts [list $_ glob]
          }
          lappend checkscripts [list \
             [$configuration GetValue script_filename_static_library] exact]
          # static library script required on all platforms

          foreach dir $libdirs {
             set dir [Oc_DirectPathname $dir]
             set candidates [list]
             foreach elt $checkscripts {
                lassign $elt script type
                set test [file join $dir [eval $script $stem]]
                set match_list [list]
                if {[string compare $type "glob"]==0} {
                   set match_list [glob -nocomplain $test]
                } else {
                   if {[file exists $test]} {
                      set match_list [list $test]
                   }
                }
                foreach match $match_list {
                   if {[lsearch -exact $candidates $match] == -1} {
                      lappend candidates $match
                   }
                }
                if {[llength $candidates]>0} {
                   lappend ret [lindex $candidates 0]
                   if {[llength $candidates]>1} {
                      # Multiple matches.  What to do?
                      puts stderr "\n*** WARNING: Multiple matches to\
                                     library stem $stem: $candidates ***\n"
                   }
                   break
                }
             }
             if {[llength $candidates]>0} {
                # Match found. Terminate library directory search and
                # process next stem.
                break
             }
          } ;# Next directory
       } ;# Next stem
       return $ret
    }

    proc AbsoluteSrc {stemlist} {
        # Assume stemlist has only one stem
        set ret {}
        lappend ret [file join [pwd] [lindex $stemlist 0]]
        return $ret
    }

    proc GetConfigValue { feature } {
       return [$configuration GetValue $feature]
    }

    proc GetIncludeDirectories {} {
       set incdirs [list]
       if {![catch {$configuration \
                       GetValue program_compiler_extra_include_dirs} chk]} {
          lappend incdirs {*}$chk
       }
       if {![catch {$configuration GetValue \
                       program_compiler_c++_includedirs_cmd} incdircmd] \
              && ![catch {{*}$incdircmd} chk]} {
          lappend incdirs {*}$chk
       }
       return $incdirs
    }

    proc GetLibraryDirectories {} {
       set libdirs [list]
       if {![catch {$configuration GetValue program_linker_extra_lib_dirs} \
                chk]} {
          lappend libdirs {*}$chk
       }
       if {![catch {$configuration GetValue \
                      program_compiler_c++_libdirs_cmd} libdircmd] \
              && ![catch {{*}$libdircmd} chk]} {
          foreach dir $chk {
             if {[file isdirectory $dir]} {
                lappend libdirs $dir
             }
          }
       }
       return $libdirs
    }

    proc FindSharedLibrary { stem version dirlist } {
       # Given a library stem, version, and a list of directories,
       # returns the full pathname of first matching shared (dynamic)
       # library.  Returns an empty string if no match found, or if the
       # platform configuration value script_filename_shared_library is
       # not set. The version should be a list of zero or more values,
       # indicating major and minor release numbers.
       if {[catch {$configuration GetValue \
                      script_filename_shared_library} shared_script] \
              || [string length $shared_script]==0 } {
          return {}
       }
       set libname [eval $shared_script $stem {*}$version]
       foreach dir $dirlist {
          set tst [join $dir $libname]
          if {[file exists $tst]} {
             return [file normalize $tst]
          }
       }
       return {}
    }

    proc LinkerLibs {stemlist} {
        set ret {}
        foreach stem $stemlist {
            if {[string match tk $stem]} {
                if {![$configuration GetValue use_tk]} { continue }
                set ret [concat $ret [$configuration GetValue TK_LIB_SPEC]]
                set ret [concat $ret [$configuration GetValue TK_LIBS]]
                continue
            }
            if {[string match tcl $stem]} {
                set ret [concat $ret [$configuration GetValue TCL_LIB_SPEC]]
                set ret [concat $ret [$configuration GetValue TCL_LIBS]]

                if {[catch {tcl::pkgconfig get libdir,runtime} libdir]} {
		   continue
                }
                if {[catch {
                   $configuration GetValue program_linker_rpath
                } rpath_script]} {
                   continue
                }
                if {![file isdirectory $libdir]} {
                    continue
                }
                if {![catch {
                   $configuration GetValue cross_compile_target_tcl_rpath
                } target_rpath] && ![catch {
                   $configuration GetValue program_linker_rpathlink
                } rpathlink_script]} {
                   # Cross-compile link with separate rpath
                   set ret [concat $ret [eval $rpath_script $target_rpath]]
                   set ret [concat $ret [eval $rpathlink_script $libdir]]
                } else {
                   # Usual link
                   set ret [concat $ret [eval $rpath_script $libdir]]
                }
                continue
            }
            if {[catch {$configuration GetValue EXTRA_LIBS_$stem} extra_libs]} {
               # Supplemental libs needed with library $stem
               set extra_libs {}
            }
            if {[$configuration GetValue program_linker_uses_-L-l]} {
               set libdir [$class FindLibraries [list $stem]]
               if {[string length $libdir]>0} {
                  set libdir [file dirname $libdir]
                  if {[file isdirectory $libdir]} {
                     # If $libdir is not a directory, then assume $stem is
                     # a system library, and so doesn't need a -L option.
                     lappend ret -L$libdir
                  }
               }
               lappend ret -l$stem
               foreach xtra $extra_libs { lappend ret -l$xtra }
            } elseif {[$configuration GetValue program_linker_uses_-I-L-l]} {
               set libdir [file dirname [$class FindLibraries [list $stem]]]
               if {[file isdirectory $libdir]} {
                  # If $libdir is not a directory, then assume $stem
                  # is a system library, and so doesn't need the -I
                  # and -L options.
                  set hdir [file dirname $libdir]
                  set header [file join $hdir $stem.h]
                  if {[file readable $header]} {
                     foreach dir [[CSourceFile New _ $header] DepPath] {
                        lappend ret -I$dir
                     }
                  }
                  lappend ret -ptr
                  lappend ret [file join $libdir tr]
                  lappend ret -L$libdir
               }
               lappend ret -l$stem
               foreach xtra $extra_libs { lappend ret -l$xtra }
            } else {
               set lf [$class FindLibraries [list $stem]]
               if {[file isfile $lf]} {
                  lappend ret $lf
               } else {
                  # Assume this is a system library
                  lappend ret $stem
                  set ret [concat $ret $extra_libs]
               }
            }
        }
        return $ret
    }

    proc Index { optlist optindex } {
       # "lindex" with guard rails, intended as a support routine for
       # CompileCmd option processing.  This routine can be referenced
       # from config/platform/<platform>.tcl. For example, suppose you
       # want
       #   Oc_Option Add * Platform cflags {-optlevel #}
       # in the config/options.tcl (or config/local/options.tcl) file to
       # return "/Od", "/Ox", or "/O2 /GS-" if "#" is <=0, ==1, or >=2,
       # respectively. Then set
       #   $config SetValue program_compiler_c++_option_optlevel {
       #      Platform Index [list /Od /Ox [list /O2 /GS-]]
       #   }
       # in <platform.tcl>.
       if {$optindex<0} {
          set optindex 0
       } elseif {$optindex >= [llength $optlist]} {
          set optindex end
       }
       return [lindex $optlist $optindex]
    }

    proc CompileCmd {language args} {
        set pfx program_compiler_[string tolower $language]
        set compcmd [$configuration GetValue $pfx]
        set args [concat $cflags $args]
        while {[string match -* [lindex $args 0]]} {
            regexp ^-(.*)$ [lindex $args 0] _ opt
            set val [lindex $args 1]
            set args [lrange $args 2 end]

            # Use the platform as an escape to pass in command line
            # stuff on a single platform only
            if {[string match $platformName $opt]} {
               foreach word $val {
                  lappend compcmd $word
               }
               continue
            } elseif {[string match -* $val]} {
               lappend suspicious_val $val ;# Record in case of error
            }

            # Pre-process the option value, if applicable
            if {[info exists compopts($opt)]} {
                #             /-- 'eval' breaks apart contents
                #             |                /- 'list' preserves argument
                #             v                v
                set val [eval $compopts($opt) [list \
                        [Platform EvalSafeList $val]]]
                #                 ^
                #                 \- Protect [eval] from any "\n" or ";"
                #                    in the value of 'val'
            }

            if {[catch {
                    $configuration GetValue ${pfx}_option_$opt
                } optScript] || ![string length $optScript]} {
                # Ignore any options not supplied by this configuration
                continue
            }
            foreach elt $val {
                set compcmd [concat $compcmd [eval $optScript [list $elt]]]
            }
        }
        if {[llength $args]} {
           set errmsg "Expected compile option, but saw '[lindex $args 0]'"
           if {[info exists suspicious_val]} {
              if {[llength $suspicious_val]>1} {
                 append errmsg " (Perhaps one of \{$suspicious_val\}\
                                got eaten as an arg?)"
              } else {
                 append errmsg " (Perhaps \"$suspicious_val\"\
                                got eaten as an arg?)"
              }
           }
           return -code error $errmsg
        }
        return $compcmd
    }

    proc Compile {language args} {
       set compcmd [eval Platform CompileCmd {$language} $args]
       if {[catch {Oc_Exec Foreground {*}$compcmd} msg]} {
          global errorCode
          return -code error -errorcode $errorCode $msg
       }
    }

    proc Link {args} {
        # Special case handling for linker -debug option: If -debug is
        # not specified in lflags, but is in cflags, then copy the
        # cflags setting into lflags. Change in behavior note: this
        # modification assumes both lflags and cflags are even sized
        # lists. This is evidently the intention from the extant
        # handling here and in proc CompileCmd, but previously either
        # flags list would allow a last, odd indexed option with an
        # implicit empty string as its value.
        set linkflags [dict create {*}$lflags]
        if {![dict exists $linkflags -debug]} {
           set compileflags [dict create {*}$cflags]
           if {[dict exists $compileflags -debug]} {
              lappend lflags -debug [dict get $compileflags -debug]
           }
        }
        set linkcmd [$configuration GetValue program_linker]
        set args [concat $lflags $args]
        set verifyargs $args
        while {[string match -* [lindex $args 0]]} {
            regexp ^-(.*)$ [lindex $args 0] _ opt
            set val [lindex $args 1]
            set args [lrange $args 2 end]

            # Use the platform as an escape to pass in command line
            # stuff on a single platform only
            if {[string match $platformName $opt]} {
                foreach word $val {
                    lappend linkcmd $word
                }
                continue
            } elseif {[string match -* $val]} {
               lappend suspicious_val $val ;# Record in case of error
            }

            # Pre-process the option value, if applicable
            if {[info exists linkopts($opt)]} {
                set val [eval $linkopts($opt) [list \
                        [Platform EvalSafeList $val]]]
            }

            if {[catch {
                    $configuration GetValue program_linker_option_$opt
                } optScript] || ![string length $optScript]} {
                # Ignore any options not supplied by this configuration
                continue
            }

            foreach elt $val {
                set linkcmd [concat $linkcmd [eval $optScript [list $elt]]]
            }
        }
        if {[llength $args]} {
           set errmsg "Expected link option, but saw '[lindex $args 0]'"
           if {[info exists suspicious_val]} {
              if {[llength $suspicious_val]>1} {
                 append errmsg " (Perhaps one of \{$suspicious_val\}\
                                got eaten as an arg?)"
              } else {
                 append errmsg " (Perhaps \"$suspicious_val\"\
                                got eaten as an arg?)"
              }
           }
           return -code error $errmsg
        }
        if {![catch {$configuration GetValue program_linker_extra_args} flags]} {
           set linkcmd [concat $linkcmd $flags]
        }

        # Workaround for ${DBGX} values sprinkled in old broken
        # tclConfig.sh and tkConfig.sh files.
        regsub -all {\${DBGX}} $linkcmd {} linkcmd
        #
        if {[catch {eval Oc_Exec Foreground $linkcmd} msg]} {
            global errorCode
            set ec $errorCode
            eval $class LinkVerify $verifyargs
            return -code error -errorcode $ec $msg
        } else {
            eval $class LinkVerify $verifyargs
        }
    }

    proc LinkVerify {args} {
        while {[string match -* [lindex $args 0]]} {
            if {[string match -out [lindex $args 0]]} {
                VerifyExecutable [lindex $args 1] $noexedelete
            }
            set args [lrange $args 2 end]
        }
    }

    proc MakeLibrary {args} {
        set mlcmd [$configuration GetValue program_libmaker]
        while {[string match -* [lindex $args 0]]} {
            regexp ^-(.*)$ [lindex $args 0] _ opt
            set val [lindex $args 1]
            set args [lrange $args 2 end]

            # Use the platform as an escape to pass in command line
            # stuff on a single platform only
            if {[string match $platformName $opt]} {
                foreach word $val {
                    lappend mlcmd $word
                }
                continue
            } elseif {[string match -* $val]} {
               lappend suspicious_val $val ;# Record in case of error
            }

            # Pre-process the option value, if applicable
            if {[info exists mlopts($opt)]} {
                set val [eval $mlopts($opt) [list \
                        [Platform EvalSafeList $val]]]
            }

            if {[catch {
                    $configuration GetValue program_libmaker_option_$opt
                } optScript] || ![string length $optScript]} {
                # Ignore any options not supplied by this configuration
                continue
            }
            foreach elt $val {
                set mlcmd [concat $mlcmd [eval $optScript [list $elt]]]
            }

        }
        if {[llength $args]} {
           set errmsg "Expected library build option, but saw '[lindex $args 0]'"
           if {[info exists suspicious_val]} {
              if {[llength $suspicious_val]>1} {
                 append errmsg " (Perhaps one of \{$suspicious_val\}\
                                got eaten as an arg?)"
              } else {
                 append errmsg " (Perhaps \"$suspicious_val\"\
                                got eaten as an arg?)"
              }
           }
           return -code error $errmsg
        }
        if {[catch {eval Oc_Exec Foreground $mlcmd} msg]} {
            global errorCode
            return -code error -errorcode $errorCode $msg
        }
    }

    proc IndexLibrary {libstem} {
        if {[catch {$configuration GetValue TCL_RANLIB} ilcmd]} {
            return
        }
        if {[string match : $ilcmd]} {
            return
        }
        set ilcmd [concat $ilcmd [Platform PlatformLibrary $libstem]]
        if {[catch {eval Oc_Exec Foreground $ilcmd} msg]} {
            global errorCode
            return -code error -errorcode $errorCode $msg
        }
    }

    Constructor {} {
        return -code error "Don't make instances of class 'Platform'"
    }
}

