# FILE: cpuguess-lintel.tcl
#
# This file defines a Tcl proc for use on Linux x86 systems,
#    GuessCpu
# to guess the CPU model.
#   Additionally, this script sources x86-support.tcl, which defines
# the following Tcl procs:
#    GuessCpuArch_VendorFamilyModel
# which returns a cpu-arch list, {vendor type sse-level} given the
# vendor, processor family and model id numbers.
#    GuessCpuArch_NameStr
# which produces a cpu-arch list like the previous proc, but using
# the "model name" string.
#    GuessGccVersion
# which runs gcc to guess the version of GCC being used.
#    GetGccGeneralOptFlags
# which returns a list of aggressive, non-processor-specific
# optimization flags that can be passed to gcc.
#    GetGccCpuOptFlags
# which returns a list of processor-specific optimization flags that
# can be passed to gcc
#    Find_SSE_Level
# which determines the SSE type level /proc/cpuinfo
#    Find_FMA_type
# which determines the FMA type from /proc/cpuinfo
#
# The GuessCpu proc is mostly a wrapper around the
# GuessCpuArch_VendorFamilyModel proc in x86-support.tcl, except that
# the SSE level is deduced directly from the /proc/cpuinfo "flags"
# line.
#   See the x86-support.tcl file for information about the other procs.
#
# This file is sourced by the lintel.tcl, linux-x86_64.tcl, cygtel.tcl,
# cpuguess.tcl, cpuguess-linux-x86_64.tcl, cpuguess-cygtel.tcl scripts.
#
########################################################################

# Support procs
source [file join [file dirname [info script]] x86-support.tcl]

# Routine to guess the CPU using /proc/cpuinfo provided by Linux kernel
# Output is a three element list.
# The first element is the vendor, which will be one of:
#     unknown, intel, or amd.
# The second element is the cpu type, which will be one of:
#     unknown,
#     i386, i486, pentium, pentium-mmx, pentiumpro, pentium2,
#     pentium3, pentium-m, core, core2
#     pentium4, prescott, nocona, pentium-d,
#     k5, k6, k6-2, k6-3, athlon, athlon-4, k8
# The third element is the supported SSE level, which will be one of:
#     0, 1, 2, 3, 4
#
# This proc is mostly a wrapper around the GuessCpuArch_VendorFamilyModel
# proc in x86-support.tcl, except that the SSE level is deduced directly
# from the /proc/cpuinfo "flags" line.

proc GuessCpu {} {
   set vendor  unknown
   set cputype unknown
   set sselevel 0

   if {[catch {
      if {[file readable /proc/cpuinfo]} {
         set fptr [open /proc/cpuinfo r]
         set cpustr [read $fptr]
         close $fptr
      } else {
         # If we are running under cygwin, then we might
         # not be able to see /proc/cpuinfo directly
         # but still be able to set cat on it.
         set cpustr [exec cat /proc/cpuinfo]
      }
      # Look at first processor info only
      set tmplst [split $cpustr "\n"]
      set record_end [lsearch $tmplst {}]
      if {$record_end>0} {
         set tmplst [lreplace $tmplst $record_end end]
      }
      set cpustr [join $tmplst ":"]
      foreach {x y} [split $cpustr ":"] {
         set cpuinfo([string trim $x]) [string tolower [string trim $y]]
      }

      if {[info exists cpuinfo(vendor_id)]} {
         set vendor $cpuinfo(vendor_id)
      } elseif {[info exists cpuinfo(model\ name)]} {
         set vendor $cpuinfo(model\ name)
      }

      set cpu_family {}
      if {[info exists cpuinfo(cpu\ family)]} {
         set cpu_family $cpuinfo(cpu\ family)
      }

      set cpu_model {}
      if {[info exists cpuinfo(model)]} {
         set cpu_model $cpuinfo(model)
      }

      foreach {vendor cputype sselevel} \
         [GuessCpuArch_VendorFamilyModel $vendor $cpu_family $cpu_model] {
            break
         }

      if {[string match "unknown" $cputype]} {
         # Otherwise, try to match against "model name"
         foreach {vendor cputype sselevel} \
            [GuessCpuArch_NameStr $cpuinfo(model\ name)] { break }
      }

      if {[info exists cpuinfo(flags)]} {
         set sselevel [Find_SSE_Level $cpuinfo(flags)]
      }
   } err]} {
      puts stderr "CODING ERROR: $err"
   }
   return [list $vendor $cputype $sselevel]
}

########################################################################
########################################################################
#
# Support code for the Intel C++ "icpc" compiler. This compiler
# has been superseded by "icpx" (see below).

# Routine to guess the Intel C++ version.  The import, icpc, is used
# via "exec $icpc --version" (or, rather, the "open" analogue) to
# determine the icpc version (since the flags accepted by icpc vary by
# version).  Return value is the icpc version string as a list of
# numbers, for example, {10 1} for version 10.1
proc GuessIcpcVersion { icpc } {
    set guess {}
    catch {
	set fptr [open "|$icpc --version" r]
	set verstr [read $fptr]
	close $fptr
        set digstr {[0-9]+\.[0-9.]+}
        set ws "\[ \t\n\r\]"
	regexp -- "(^|$ws)($digstr)($ws|$)" $verstr dummy dum0 guess dum1
    }
    return [split $guess "."]
}

proc GetIcpcBannerVersion { icpc } {
   set banner {}
   catch {
      set fptr [open "|$icpc --version" r]
      set banner [gets $fptr]
      close $fptr
   }
   return $banner
}

# Routines that report optimization flags for icpc.  Right now these
# are mostly placeholders, but they may be expanded and refined in the
# future.
proc GetIcpcGeneralOptFlags { icpc_version } {
   set opts [list -xHost -O3 -ipo -no-prec-div -ansi_alias \
                -fp-model fast=2 -fp-speculation fast \
                -std=c++11]
   return $opts
}
proc GetIcpcValueSafeOptFlags { icpc_version } {
   set opts [list -xHost -O3 -ipo -ansi_alias \
                -fp-model precise -no-ftz -fp-speculation fast \
                -std=c++11]
   # AFAICT, the -no-ftz (no flush to zero) flag doesn't hurt, but
   # is only operative when compiling main().  All other routines
   # inherit the FPU settings from the caller.
   return $opts
}
proc GetIcpcCpuOptFlags { icpc_version cpu_arch } {
   # CPU model architecture specific options.
   set cpuopts {}
   set cpu_vendor "unknown"
   set cpu_type   "unknown"
   set cpu_sse    0
   foreach {cpu_vendor cpu_type cpu_sse} $cpu_arch {
      set cpu_vendor [string tolower $cpu_vendor]
      set cpu_type   [string tolower $cpu_type]
      break
   }
   if {[lindex $icpc_version 0]<11} {
      # The following flags are based on the docs for icc version 10.0.
      # Allowed flags may differ for other versions of icc.  Note: -xT
      # and -xP imply SSE3.
      switch -exact -- $cpu_type {
         athlon-4  { lappend cpuopts -xK }
         opteron   -
         k8        { lappend cpuopts -xW}
         pentium3  { lappend cpuopts -xK }
         pentium-m { lappend cpuopts -xW}
         pentium4  { lappend cpuopts -xN}
         prescott  -
         nocona    -
         pentium-d -
         core      { lappend cpuopts -xP }
         core2     { lappend cpuopts -xT }
         default   {
            if {$cpu_sse>=3 && [string match intel $cpu_vendor]} {
               lappend cpuopts -xP
            } elseif {$cpu_sse>=2} {
               lappend cpuopts -xW
            } elseif {$cpu_sse>=1} {
               lappend cpuopts -xK
            }
         }
      }
   } else {
      # The following flags are based on the docs for icc version 11.
      switch -exact -- $cpu_type {
         athlon-4  { lappend cpuopts -xSSE }
         opteron   -
         k8        { lappend cpuopts -msse2}
         pentium3  { lappend cpuopts -xSSE }
         pentium-m { lappend cpuopts -msse2}
         pentium4  { lappend cpuopts -xSSE2}
         prescott  -
         nocona    -
         pentium-d -
         core      { lappend cpuopts -xSSE3}
         core2     { lappend cpuopts -xSSSE3}
         default   {
            if {$cpu_sse>=3 && [string match intel $cpu_vendor]} {
               lappend cpuopts -xSSE3
            } elseif {$cpu_sse>=2} {
               lappend cpuopts -msse2
            } elseif {$cpu_sse>=1} {
               lappend cpuopts -xSSE
            }
         }
      }
   }

   return $cpuopts
}

########################################################################
########################################################################
#
# Support code for the Intel C++ "icpx" compiler. This is Intel's
# replacement for the older "icpc" compiler (see above).
proc GuessIcpxVersion { icpx } {
    set guess {}
    catch {
	set fptr [open "|$icpx --version" r]
	set verstr [read $fptr]
	close $fptr
        set digstr {[0-9]+\.[0-9.]+}
        set ws "\[ \t\n\r\]"
	regexp -- "(^|$ws)($digstr)($ws|$)" $verstr dummy dum0 guess dum1
    }
    return [split $guess "."]
}

proc GetIcpxBannerVersion { icpx } {
   set banner {}
   catch {
      set fptr [open "|$icpx --version" r]
      set banner [gets $fptr]
      close $fptr
   }
   return $banner
}

proc GetIcpxIncludeDirs { icpx } {
   # This code is based off of proc GetClangIncludeDirs in
   # clang-support.tcl. This is nearly the same as GetGccIncludeDirs in
   # gcc-support.tcl, and icpx seems to support the gcc syntax as well.
   # Tested against
   #    Intel(R) oneAPI DPC++/C++ Compiler 2024.0.0 (2024.0.0.20231017)
   set testcmd [concat | $icpx -Wp,-v -stdlib=libc++ -x c++ - \
                   -fsyntax-only < /dev/null 2>@1]
   set incdirs [list]
   catch {
      set fptr [open $testcmd r]
      set output [split [read $fptr] "\n"]
      close $fptr
      foreach line $output {
         if {[regexp {^(#|clang |End of search|ignoring)} $line]} {
            continue
         }
         if {[regexp {^\s*(/.*\S)\s*$} $line _ dir] \
                && [lsearch -exact $incdirs $dir]<0 \
                && [file isdirectory $dir]} {
            lappend incdirs $dir
         }
      }
   }
   return $incdirs
}

proc GetIcpxLibDirs { icpx } {
   set testcmd [concat | $icpx -print-search-dirs 2>@1]
   set libdirs [list]
   catch {
      set fptr [open $testcmd r]
      set output [split [read $fptr] "\n"]
      close $fptr
      foreach line $output {
         if {[regexp {^libraries:\s*=?(.*)$} $line _ dirs]} {
            lappend libdirs {*}[split $dirs :]
         }
      }
   }
   return $libdirs
}

# Routines that report optimization flags for icpx.
proc GetIcpxGeneralOptFlags { icpx_version optlevel } {
   set opts [list]
   # For reference, the C++ standards level is set using
   # e.g. -std=c++20.  Default is c++17.
   set opts "-std=c++11"
   if {$optlevel<3} {
      lappend opts -O$optlevel
   } else {
      lappend opts -fast
   }
   return $opts
}
proc GetIcpxValueSafeOptFlags { icpx_version optlevel } {
   set opts [list]
   # Set C++ standards level using, e.g., /Qstd:c++17.
   # Default is c++11, but if you include that the compiler
   # complains "argument unused during compilation"
   set opts "-std=c++11"
   if {$optlevel<3} {
      lappend opts -O$optlevel
   } else {
      lappend opts -O3 -ipo
   }
   lappend opts -fp-model=precise
   return $opts
}
proc GetIcpxCpuOptFlags { icpx_version cpu_arch } {
   # CPU model architecture specific options.
   set cpuopts {}
   return $cpuopts
}


########################################################################
########################################################################


# Routines to obtain version from Portland Group pgc++ compiler.
proc GuessPgccVersion { pgcc } {
   set guess {}
   catch {
      set fptr [open "|$pgcc --version" r]
      set verstr [string trim [read $fptr]]
      close $fptr
      set verstr [lindex [split $verstr "\n"] 0]
      set digstr {[0-9]+\.[0-9.-]+}
      set ws "\[ \t\n\r\]"
      regexp -- "(^|$ws)($digstr)($ws|$)" $verstr dummy dum0 guess dum1
   }
   return [split $guess ".-"]
}

proc GetPgccBannerVersion { pgcc } {
   set banner {}
   catch {
      set fptr [open "|$pgcc --version" r]
      set verstr [string trim [read $fptr]]
      close $fptr
      set banner [lindex [split $verstr "\n"] 0]
   }
   return $banner
}

proc GetPgccGeneralOptFlags { pgcc_version } {
   # Note: If flags -mp and --exceptions are both used, the latter must
   # come later in the option sequence.
   set opts [list -std=c++11]
   if {[lindex $pgcc_version 0]<=16} {
       lappend opts --exceptions
   }
   return $opts
}
proc GetPgccValueSafeOptFlags { pgcc_version cpu_arch} {
   set opts  [list -O2 -Kieee -Mvect=noassoc,fuse,prefetch \
                 -Mcache_align -Mprefetch -Msmartalloc -Mnoframe \
                 -Munroll -Mnoautoinline]
   return $opts
}
proc GetPgccFastOptFlags { pgcc_version cpu_arch } {
   # CPU model architecture specific options.
   set opts [list -fast -Knoieee -Mnoautoinline]
   return $opts
}

proc GetPgccLinkOptFlags { pgcc_version } {
   # Linker flags
   set opts {-Mnoeh_frame}
   ## Without -Mnoen_frame, threads in Oxs abort during
   ## Tcl_ExitThread() processing with the error "terminate called
   ## without an active exception" if the thread code contains a
   ## try/catch block.  (At least for pgc++ version 16.10-0.)
   if {[lindex $pgcc_version 0]<=16} {
       lappend opts -latomic
   }
   return $opts
}
