# FILE: makerules.tcl
#
# This file controls the application 'pimake' by defining rules which
# describe how to build the applications and/or extensions produced by
# the source code in this directory.
#
# Verify that this script is being sourced by pimake
if {[llength [info commands MakeRule]] == 0} {
    error "'[info script]' must be evaluated by pimake"
}
#
# Notes on LaTeX .aux files:
#
# Both latex and pdflatex create and use .aux files when run.  Both will
# fail if run with an existing .aux files from the other. This would be
# a particular issue if latex and pdflatex were run in parallel with
# both writing into the same directory. Also, latex2html uses these .aux
# files. Versions 2018.2 and earlier require latex-style .aux
# files. Versions 2018.3 and later expect pdflatex-style .aux file
# unless the -nouse_pdftex option is specified. Beware too that the
# various TeX-ish programs support different command line options and
# behavior between the Linux TeX Live and Windows MiKTeX
# distributions. The approach implemented in this makerules.tcl script
# is
#
#  (0) LaTeX2HTML 2018.2 and earlier versions use use latex, not
#      pdflatex.  The default latex2html-init HTML_VERSION setting is
#      4.0 in LaTeX2HTML 2018.2; this seems to be the case at least as
#      far back as LaTeX2HTML 2015. For LaTeX2HTML 2018.3 and latter
#      latex pdflatex is used by default instead of latex, and the
#      default latex2html-init HTML_VERSION setting is 5.0 through at
#      least verston 2024.
#
#  (1) All .tex files live in the toplevel progman directory.
#
#  (2) LaTeX2HTML uses progman/progman as its workspace and final
#      destination. The required .aux files are written here as well.
#      latex2html requires the -rootdir option to find the .aux files
#      when they are not next to the .tex sources. (There is an
#      "-external_file" option to latex2html that allows the user to
#      specify a specific .aux file, but AFAICT only one file may be
#      specified this way, and latex and pdflatex drop separate .aux
#      files for each section.)
#
#  (3) For LaTeX2HTML 2018.2 and earlier:
#
#    (3.1) progman/progman is used as the workspace for latex .dvi. The
#      progman.dvi build script moves the completed progman.dvi up to
#      progman/, which is where dvips is run to build the :ps target.
#
#    (3.2) The subdirectory progman/progman-pdf/ is used as the build
#      area for :pdflatex, and when complete progman.pdf is moved from
#      progman/progman-pdf/ to progman/.
#
#    (3.3) progman/progman-dvi is not created.

#  (4) For LaTeX2HTML 2018.3 and later:
#
#    (4.1) progman/progman is used as the workspace for pdflatex. The
#      build script moves the completed progman.pdf up to progman/.
#
#    (4.2) The subdirectory progman/progman-dvi/ is used as the build
#      area for progman.dvi, but when progman.dvi is complete the build
#      script moves it up to progman/. The dvips program is run in
#      progman/ to build the :ps target.
#
#    (4.3) progman/progman-pdf is not created.
#
########################################################################
#
# Phony targets: doc maintainer-clean distclean clean upgrade
#                 :pdf :ps :latex2html :latexml

# MAKERULES.TCL scripting notes:
#
# (1) Each makerules.tcl script is run in its own namespace; this
# protects makerules.tcl scripts from stepping on one another. In
# particular, if you define a proc inside a makerules.tcl script and
# want to alias it inside a child interpreter, then you need to resolve
# the namespace in the assignment, like so:
#
#   proc Foo {} { puts Hi }
#   interp alias $slave Bar {} [namespace current]::Foo
#
# (2) The MakeRule -script argument is passed off to proc Build in
# app/pimake/makerule.tcl, where variables defined in the makerules.tcl
# script aren't directly in scope. The simplest workaround is usually to
# wrap the -script argument inside a [subst -nocommands { ... }],
# although beware of embedded spaces and other abnormalities in the
# interpolated variable value. Also remember to backslash quote any '$'
# that are meant to be interpolated inside the script evaluation, e.g.,
# loop variables.
#
# (3) Variables defined in makerules.tcl scripts are typically deleted
# at the end of the script, as they are in this script.  So any variable
# values that are needed following the end of the evaluation of a
# makerules.tcl need to be retained separately, i.e., you can't access
# variable [namespace current]::fubar after the end of its makerules.tcl
# file evaluation.

# Make all
MakeRule Define {
   -targets            doc
   -dependencies       [list :pdf :ps :latex2html :latexml]
}
## Note: :latex2html requires .aux from a pdflatex or latex run
##       (depending on the latex2html version), so it is maybe best if
##       :latex2html follows :pdf and :ps on the above dependencies
##       list.

# BibTeX command: The Windows MiKTeX-BibTeX 4.1 (MiKTeX 23.4) version of
# bibtex doesn't include the auxfile directory in the search path, so
# for example
#    bibtex userguide-pdf/userguide
# chokes on lines in userguide-pdf/userguide that look like
#    \@input{overview-body.aux}
# where overview-body.aux lives in userguide/userguide-pdf/ rather than
# next to userguide.tex in userguide/. We can work around this by using
# the -include-directory option to MiKTeX-BibTeX, but this option is not
# supported in the TeX Live 2022 version of bibtex on Linux. So provide
# an installation specific workaround:
proc GetBibtexCmd { dir } {
   if {![info exists use_bibtex_incdir]} { ;# Check for cached value
      set tmpchk 0
      if {[string compare $::tcl_platform(platform) windows] == 0} {
         # Running on Windows.
         if {![catch {eval exec bibtex -help} _]} {
            # Note: Use exec rather than "Oc_Exec Foreground" because
            # the latter forces redirection of stdout instead of
            # capturing it.
            if {[regexp -lineanchor -- {^\s+-include-directory} $_]} {
               set tmpchk 1
            }
         }
      }
      set use_bibtex_incdir $tmpchk
   }
   if {$use_bibtex_incdir} {
      return [list bibtex -include-directory $dir]
   }
   return bibtex
}

# LaTeX2HTML version-specific behavior
# Shortcut target for latex2html build. Check existence and version.
# Note: Whitespace in the *auxdir variable values will break
#    interpolation in MakeRule '-script [subst ...]' code below.
#    avoid or adjust accordingly.
set dviauxdir userguide-dvi
set pdfauxdir userguide-pdf
if {![llength [auto_execok latex2html]] \
       || ![regexp {LaTeX2HTML Version +([0-9]+)} \
               [exec -ignorestderr latex2html -version 2>@1] \
               _ l2hversion] \
       || $l2hversion < 2018} {
   # Bad or missing latex2html
   set good_latex2html 0
} else {
   # Presumably good latex2html
   set good_latex2html 1
   if { $l2hversion < 2019 } {
      # LaTeX2HTML versions prior to 2019 require latex, not pdflatex
      set dviauxdir userguide
      set l2hdeps [list userguide.dvi]
   } else {
      set pdfauxdir userguide
      set l2hdeps [list userguide.pdf]
   }
}


# PDF target. Note: Two passes of pdflatex may be needed after makeindex
# to get Index included in the table of contents. Also, on at least some
# platforms if the subdirectory "userguide" (which holds latex2html
# output) exists, 'makeindex userguide' will try to use the subdirectory
# instead of userguide.idx. So it is important to include the .idx
# suffix in the makeindex call.
MakeRule Define {
   -targets            userguide.pdf
   -dependencies       [glob -nocomplain *.tex *.bib \
                           [file join .. common *.tex]]
   -script [subst -nocommand {
      DeleteFiles $pdfauxdir
      MakeDirectory $pdfauxdir
      foreach fn [glob -nocomplain *.tex] {
         set fn [regsub {tex$} \$fn aux]
         close [open [file join $pdfauxdir \$fn] a]
      }
      Oc_Exec Foreground pdflatex -output-directory=$pdfauxdir userguide.tex
      Oc_Exec Foreground {*}[GetBibtexCmd $pdfauxdir] \
         [file join $pdfauxdir userguide]
      Oc_Exec Foreground pdflatex -output-directory=$pdfauxdir userguide.tex
      Oc_Exec Foreground makeindex [file join $pdfauxdir userguide.idx]
      Oc_Exec Foreground pdflatex -output-directory=$pdfauxdir userguide.tex
      Oc_Exec Foreground pdflatex -output-directory=$pdfauxdir userguide.tex
      RenameFile [file join $pdfauxdir userguide.pdf] userguide.pdf
   }]
}

MakeRule Define {
   -targets         :pdf
   -dependencies    userguide.pdf
}


# Postscript target. Note: Two passes of latex may be needed after
# makeindex to get Index included in the table of contents. Also, on at
# least some platforms if the subdirectory "userguide" (which holds
# latex2html output) exists, 'makeindex userguide' will try to use the
# subdirectory instead of userguide.idx. So it is important to include
# the .idx suffix in the makeindex call.
MakeRule Define {
   -targets            userguide.ps
   -dependencies       userguide.dvi
   -script      {
      DeleteFiles userguide.ps
      Oc_Exec Foreground dvips -Ppdf -f userguide.dvi > userguide.ps
   }
}

MakeRule Define {
   -targets            userguide.dvi
   -dependencies       [glob -nocomplain *.tex *.bib \
                           [file join .. common *.tex]]
   -script        [subst -nocommand {
      # An alternative to complete deletion of userguide/ is
      # to delete only userguide/*.{aux,bbl,ind,toc}
      DeleteFiles $dviauxdir
      MakeDirectory $dviauxdir
      Oc_Exec Foreground latex -output-directory=$dviauxdir userguide.tex
      Oc_Exec Foreground {*}[GetBibtexCmd $dviauxdir] \
         [file join $dviauxdir userguide]
      Oc_Exec Foreground latex -output-directory=$dviauxdir userguide.tex
      Oc_Exec Foreground makeindex [file join $dviauxdir userguide.idx]
      Oc_Exec Foreground latex -output-directory=$dviauxdir userguide.tex
      Oc_Exec Foreground latex -output-directory=$dviauxdir userguide.tex
      RenameFile [file join $dviauxdir userguide.dvi] userguide.dvi
   }]
}

MakeRule Define {
   -targets         :ps
   -dependencies    userguide.ps
}



# LaTeX2HTML target
if {!$good_latex2html} {
   # LaTeX2HTML version not supported
   MakeRule Define {
      -targets         :latex2html
      -dependencies    {}
      -script {
         puts stderr \
            "\n*** Skipping target :latex2html\
             (bad or missing executable) ***\n"
      }
   }
} else {
   # Supported LaTeX2HTML version
   MakeRule Define {
      -targets         [file join userguide userguide.html]
      -dependencies    $l2hdeps
      -script      [subst -nocommands {
         # Note: latex2html needs the *.aux files as created by
         # pdflatex or latex (depending on the latex2html version) and
         # the associated userguide.bbl from bibtex. Throw an error if
         # these appear to be missing.
         if {![file exists [file join userguide userguide.aux]] \
                || ![file exists [file join userguide userguide.bbl]]} {
            puts stdout \
               "\n*** Skipping target :latex2html\n*** --> Missing LaTeX\
               .aux files; run 'pimake maintainer-clean' and retry\n"
            ### NB: For some reason stderr output, but not stdout, is
            ### lost in parallel, but not serial, pimake builds.
            error ":latex2html target skipped --- LaTeX .aux files missing"
         }

         # Run latex2html
         DeleteFiles {*}[glob -nocomplain [file join userguide img*]]
         set destimgdir [file join userguide images]
         file delete -force -- \$destimgdir
         Oc_Exec Foreground latex2html -init_file latex2html-init \
            -rootdir userguide userguide.tex
         file mkdir \$destimgdir
         foreach f [glob -nocomplain -directory images *.png *.svg] {
            file copy -force \$f \$destimgdir
         }
         file copy -force [file join .. common images oommficon.png] \$destimgdir
         file copy -force [file join .. common images oommficon.svg] \$destimgdir
         unset destimgdir f
      }]
   }
   MakeRule Define {
      -targets         :latex2html
      -dependencies    [file join userguide userguide.html]
   }
}


# LaTeXML target
# Rules for latexml .tex -> .xml -> .html processing.
# Unlike latex2html, which has side dependencies on the .bbl and .aux
# files from the latex build, the only dependencies for latexml are the
# *.tex files.
MakeRule Define {
   -targets            userguide.xml
   -dependencies       [glob -nocomplain *.tex [file join .. common *.tex] \
                           [file join .. common xmlextras *.ltxml]]
   -script      {
       Oc_Exec Foreground latexml \
          --path=../common/xmlextras --dest=userguide.xml userguide
   }
}

MakeRule Define {
   -targets            oommf.bib.xml
   -dependencies       oommf.bib
   -script      {
      Oc_Exec Foreground latexml --dest=oommf.bib.xml oommf.bib
   }
}

MakeRule Define {
   -targets        [file join userguide-xml userguide.html]
   -dependencies   [concat  userguide.xml oommf.bib.xml \
                       [glob -nocomplain \
                           [file join .. common xmlextras css *.css] \
                           [file join .. common xmlextras js *.js] \
                           [file join .. common xmlextras xslt *.xsl]]]
   -script     {
       Oc_Exec Foreground latexmlpost \
         --split --splitat=subsection --splitnaming=label\
         --urlstyle=file \
         --nodefaultresources \
         --css=../common/xmlextras/css/oommf.css \
         --css=../common/xmlextras/css/oommf-report.css \
         --css=../common/xmlextras/css/oommf-tweakfonts.css \
         --navigationtoc=context \
         --css=../common/xmlextras/css/oommf-navbar.css \
         --javascript=../common/xmlextras/js/LaTeXML-maybeMathjax.js \
         --timestamp=0 \
         --stylesheet=../common/xmlextras/xslt/oommf-html5.xsl \
         --dest=userguide-xml/userguide.html userguide.xml
   }
}

# Shortcut target for latexml-based html build
if {[llength [auto_execok latexml]]} {
   MakeRule Define {
      -targets            :latexml
      -dependencies       [file join userguide-xml userguide.html]
   }
} else {
   MakeRule Define {
      -targets            :latexml
      -dependencies       {}
      -script {
         puts stderr "\n*** Skipping target :latexml (missing executable) ***\n"
      }
   }
}

# Maintainer-clean target deletes all files except the base set which
# can be used to reconstruct everything. This should leave only files
# that are in the repo.
MakeRule Define {
   -targets            maintainer-clean
   -dependencies       distclean
   -script              {
      DeleteFiles userguide userguide.dvi userguide.ps userguide.pdf
      DeleteFiles userguide-xml userguide.xml oommf.bib.xml
   }
}

# Distclean target removes all files not needed by end user.
MakeRule Define {
   -targets             distclean
   -dependencies        clean
   -script              {
      foreach sfx {aux bbl dvi xml} {
         DeleteFiles {*}[glob -nocomplain [file join userguide *.$sfx]]
      }
   }
}

# Clean target removes all files not needed by top-level
# targets (make all).
MakeRule Define {
   -targets            clean
   -dependencies       {}
   -script             {
      # Keep *.{aux,bbl,dvi} for latex2html
      foreach sfx {blg idx ilg ind log out toc} {
         DeleteFiles {*}[glob -nocomplain *.$sfx]
         DeleteFiles {*}[glob -nocomplain [file join userguide *.$sfx]]
      }
      unset sfx
      DeleteFiles {*}[glob -nocomplain [file join userguide images.*]]
      DeleteFiles {*}[glob -nocomplain [file join userguide-xml *.cache]]
      DeleteFiles userguide-pdf
      DeleteFiles userguide-dvi
   }
}

# WARNING: upgrade target has not been maintained for a very long time...
### MakeRule Define {
###    -targets            upgrade
###    -script             {
###       DeleteFiles {*}[glob -nocomplain *.html *.css *.gif *.png *.pl]
###       DeleteFiles images.tex
###       DeleteFiles batchsys-body.tex
###       DeleteFiles batchsys.tex
###       DeleteFiles mmsolve2d-body.tex
###       DeleteFiles mmsolve2d.tex
###       DeleteFiles oxsii-body.tex
###       DeleteFiles oxsii.tex
###       DeleteFiles images.tex
###       DeleteFiles [file join userguide 2D_Micromagnetic_Solver_mmS.html]
###       DeleteFiles [file join userguide Batch_Scheduling_System.html]
###       DeleteFiles [file join userguide Data_table_format_ODT.html]
###       DeleteFiles [file join userguide OOMMF_Batch_System.html]
###       DeleteFiles [file join userguide Problem_specification_forma.html]
###       DeleteFiles [file join userguide Solver_Batch_Interface_batc.html]
###       DeleteFiles [file join userguide fileformats.html]
###       DeleteFiles [file join userguide images.pl]
###       DeleteFiles [file join userguide vectorfieldformat.html]
###    }
### }
MakeRule Define {
   -targets            upgrade
   -script             {
      error "Userguide upgrade target not currently supported."
   }
}

# NOTES:
# latexmlpost format options:
#
#   --destination=destination ... Specifies the destination file and
#      directory. The directory is needed for mathimages, mathsvg and
#      graphics processing.
#
#   --format=(html|html5|html4|xhtml|xml) ... Specifies the output
#      format for post processing. By default, it will be guessed from
#      the file extension of the destination (if given), with html
#      implying html5, xhtml implying xhtml and the default being xml,
#      which you probably don’t want.
#
#   --omitdoctype, --noomitdoctype ... Omits (or includes) the document
#      type declaration. The default is to include it if the document
#      model was based on a DTD.
#
#   --numbersections, --nonumbersections ... Includes (default), or
#      disables the inclusion of section, equation, etc, numbers in
#      the formatted document and crossreference links.
#
#   --css=cssfile ............ Adds cssfile as a css stylesheet to be
#       used in the transformed html/html5/xhtml. Multiple stylesheets
#       can be used; they are included in the html in the order given,
#       following the default ltx-LaTeXML.css. The stylesheet is copied
#       to the destination directory, unless it is an absolute url.
#
#       Some stylesheets included in the distribution are
#
#          –css=navbar-left Puts a navigation bar on the left. (default
#               omits navbar) –css=navbar-right Puts a navigation bar on
#               the left.
#          –css=theme-blue A blue coloring theme for headings.
#          –css=amsart A style suitable for journal articles.
#
#   --stylesheet=xslfile ..... Requests the XSL transformation of the
#      document using the given xslfile as stylesheet. If the stylesheet
#      is omitted, a ‘standard’ one appropriate for the format (html4,
#      html5 or xhtml) will be used.
#
#   --xsltparameter=name:value ... Passes parameters to the XSLT
#      stylesheet. See the manual or the stylesheet itself for available
#      parameters.
#
#   --timestamp=timestamp .... Provides a timestamp (typically a time
#      and date) to be embedded in the comments by the stock XSLT
#      stylesheets. If you don’t supply a timestamp, the current time
#      and date will be used. (You can use --timestamp=0 to omit the
#      timestamp).
#
#   --icon=iconfile .......... Copies iconfile to the destination
#      directory and sets up the linkage in the transformed
#      html/html5/xhtml to use that as the ”favicon”.
#
#   --javascript=jsfile ...... Includes a link to the javascript file
#      jsfile, to be used in the transformed html/html5/xhtml. Multiple
#      javascript files can be included; they are linked in the html in
#      the order given. The javascript file is copied to the destination
#      directory, unless it is an absolute url.
#
#   --nodefaultresources ..... Disables the copying and inclusion of
#      resources added by the binding files; This includes CSS,
#      javascript or other files. This does not affect resources
#      explicitly requested by the --css or --javascript options.

unset -nocomplain l2hversion l2hdeps dviauxdir pdfauxdir \
   good_latex2html use_bibtex_incdir
