/* FILE: tclobjarray.h          -*-Mode: c++-*-
 *
 * Wrapper for Tcl_Obj objects.  Inert if building against Tcl 8.0
 * or earlier.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2011/10/07 23:23:08 $
 * Last modified by: $Author: donahue $
 */

#ifndef _NB_TCLOBJARRAY
#define _NB_TCLOBJARRAY

#include <string>
#include "oc.h"

OC_USE_STRING;         // Map String --> std::string

/* End includes */

////////////////////////////////////////////////////////////////////////
// Wrapper for Tcl_Obj*.
// This wrapper implements Tcl_Obj reference counting behavior in a
// manner that allows Nb_TclObjWrapper objects to be used in C++
// standard library containers.
//
// When given a non-null Tcl_Obj*, the Nb_TclObjWrapper constructors
// store the pointer and increment the object's refCount. The
// destructor and Free() members decrement the stored Tcl_Obj*'s
// refCount. The Set() and operator=() members increment the refCount
// of the incoming Tcl_Obj and decrement the refCount on the Tcl_Obj
// being released. The Tcl library is responsible for deleting the
// Tcl_Obj when the refCount changes from 1 to 0.
//
// Clients of the Nb_TclObjWrapper Get() member are responsible for
// adjusting the refCount on the exported Tcl_Obj* consistent with
// normal Tcl_Obj semantics, i.e.,the client must increment the
// refCount if they want the object to persist beyond future
// operations on the Nb_TclObjWrapper object. In that case the client
// is responsible for ensuring the refCount is decremented at the
// appropriate time, as otherwise the code will leak memory.
//
struct Nb_TclObjWrapper {
private:
  Tcl_Obj* obj;
public:
  void Set(Tcl_Obj* other_obj) {
    if(obj) Tcl_DecrRefCount(obj);
    obj = other_obj;
    if(obj) Tcl_IncrRefCount(obj);
  }
  Tcl_Obj* Get() { return obj; }
  void Free() {
    if(obj) Tcl_DecrRefCount(obj);
    obj = nullptr;
  }

  Nb_TclObjWrapper() : obj(nullptr) {}
  Nb_TclObjWrapper(Tcl_Obj* other) : obj(other) {
    if(obj) Tcl_IncrRefCount(obj);
  }
  Nb_TclObjWrapper(const Nb_TclObjWrapper& other) : obj(other.obj) {
    if(obj) Tcl_IncrRefCount(obj);
  }
  Nb_TclObjWrapper& operator=(const Nb_TclObjWrapper& other) {
    Set(other.obj);
    return *this;
  }

  ~Nb_TclObjWrapper() {
    if(obj) Tcl_DecrRefCount(obj);
  }
};


////////////////////////////////////////////////////////////////////////
// Wrapper for Tcl_Obj**
typedef Tcl_Obj* NbTclObjPtr;
class Nb_TclObjArray
{
private:
  int size;
  Tcl_Obj** arr;
  void Alloc(int arrsize);
  void Free();
public:
  Nb_TclObjArray() : size(0), arr(NULL) {}
  Nb_TclObjArray(int arrsize) : size(0), arr(NULL) { Alloc(arrsize); }
  ~Nb_TclObjArray() { Free(); }

  void Resize(int newsize) {
    if(newsize != size) {
      Free(); Alloc(newsize);
    }
  }

  NbTclObjPtr& operator[](int index) {
#ifndef NDEBUG
    if(index>=size || index<0) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_TclObjArray",
                            "operator[]",1024,
                            "Array out-of-bounds; index=%d, size=%d",
                            index,size));
    }
#endif
    return arr[index];
  }

  const NbTclObjPtr& operator[](int index) const {
#ifndef NDEBUG
    if(index>=size || index<0) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_TclObjArray",
                            "operator[] const",1024,
                            "Array out-of-bounds; index=%d, size=%d",
                            index,size));
    }
#endif
    return arr[index];
  }

  int Size() const { return size; }
  Tcl_Obj** Array() { return arr; }
  Tcl_Obj* const * Array() const { return arr; }

  // Write interface routines.  These handle the
  // copy-on-write semantics of the Tcl referencing
  // system.
  void WriteString(int index,const char*);
  void WriteString(int index,const String& str);
  void WriteDouble(int index,double val);
  void WriteInt(int index,OC_INT4m val);
  void WriteInt(int index,OC_INT8m val);
  void WriteList(int index,const Nb_TclObjArray& val);
  /// Add more as needed

  void ReplaceObj(int index,NbTclObjPtr val);
  /// Increments refcount on val, decrements refcount on
  /// current obj, and replaces current obj with val.

  // Read interface routines.
  String GetString(int index) const;

};


#endif // _NB_TCLOBJARRAY
