/* FILE: colorquantity.h                 -*-Mode: c++-*-
 *
 * Color quantities for use by Vf_Mesh objects.
 *
 */

#ifndef _VF_COLORQUANTITY
#define _VF_COLORQUANTITY

#include <list>
#include <iostream>
#include <exception>
#include <string>
#include <unordered_map>

#include "mesh.h"   // Vf_DisplayVector::SHADEVEC_T
#include "nb.h"
#include "ocport.h"

/* End includes */     /* Optional directive to build.tcl */

#ifndef DO_DEBUG
# define DO_DEBUG 0
#endif

class Vf_ColorQuantity {
public:
  using ColorQuantityFcn = void (*)
    (const Nb_Vec3<OC_REAL4>& v,
     Vf_DisplayVector::SHADEVEC_T& shade);

  enum class EdgeType { Trunc, Periodic };
  // Note: No using-enum-declarations before C++20

  struct ColorQuantityAttributes {
  public:
    ColorQuantityFcn fcn;
    EdgeType edge;
    ColorQuantityAttributes() : fcn(nullptr), edge(EdgeType::Trunc) {}
    ColorQuantityAttributes(ColorQuantityFcn fcn_in,
                            EdgeType edge_in = EdgeType::Trunc)
      : fcn(fcn_in), edge(edge_in) {}

  };

  Vf_ColorQuantity() = delete;
  Vf_ColorQuantity(const char* name_,ColorQuantityAttributes attrib)
    : name(name_) {
    if(!registered_quantities.insert({name,attrib}).second) {
      // NB: This constructor is used for static initialization, and
      // exceptions can't be thrown during static initialization. Try
      // to at least print a helpful error message.
      std::cerr << "Error in Vf_ColorQuantity registration: name "
                << name_ << " already registered." << std::endl;
      std::terminate();
    }
    registered_names.push_back(name);
  }
  ~Vf_ColorQuantity() {
    registered_quantities.erase(name);
    registered_names.remove(name); // NB: The .remove operation
    /// removes all instances of name. This can be wasteful if the
    /// list is long, since by construction each name is unique and
    /// thus the operation could stop once a match is found. But
    /// this list should be relatively short, and checking for all
    /// instances is arguably more robust.
  }

  static ColorQuantityAttributes
  GetAttributes(const char* name) {
    ColorQuantityAttributes attrib; // Default initialization
    try {
      if(name) {
        attrib = registered_quantities.at(name);
      }
    } catch(std::out_of_range&) {
      // Keep default attrib initialization
    }
    return attrib;
  }

  static std::vector<std::string> GetNames() {
    std::vector<std::string> quantnames;
    for(const std::string& str : registered_names) {
      quantnames.push_back(str);
    }
    return quantnames;
  }

#if DO_DEBUG
  static void PrintNames() {
    int index = 0;
    for(const std::string& name : registered_names) {
      std::cout << "Name[" << index++ << "] : "
                << name << std::endl;
    }
  }
#endif // DO_DEBUG

private:
  const std::string name; // Needed for delete operation
  static std::unordered_map<std::string,
                            ColorQuantityAttributes> registered_quantities;
  static std::list<std::string> registered_names;
};


#endif // _VF_COLORQUANTITY
